#!/bin/bash

[[ ! -d bin ]] && mkdir bin

if [[ ! -f bin/s2p2bin_plus ]]; then
	echo "Compiling s2p2bin_plus..."
	g++ -O3 -s -Ibuild_source -o bin/s2p2bin_plus build_source/s2p2bin_plus.cpp build_source/saxman.cpp &> /dev/null
fi

if [[ ! -f bin/fixpointer ]]; then
	echo "Compiling fixpointer..."
	g++ -O3 -s -o bin/fixpointer build_source/fixpointer.cpp &> /dev/null
fi

if [[ ! -f bin/fixheader ]]; then
	echo "Compiling fixheader..."
	g++ -O3 -s -o bin/fixheader build_source/fixheader.cpp &> /dev/null
fi

[[ -f KEH.bin ]] && mv -f KEH.bin KEH.prev.bin
rm -f s2.p s2.h s2.log

debug_syms=""
print_err="-E -q"
listing_opt="-L"
multipass_warn="-r 2"

for n in `seq 1 2`; do
	if [[ "$1" == "-ds" ]]; then
		debug_syms="-g MAP"
		echo "Will generate debug symbols"
	elif [[ "$1" == "-pe" ]]; then
		print_err=""
		echo "Selected detailed assembler output"
	elif [[ "$1" == "-nl" ]]; then
		listing_opt=""
		echo "Listing generation disabled"
	elif [[ "$1" == "-nm" ]]; then
		listing_opt=""
		echo "Multipass warnings disabled"
	fi
	shift
done

echo Assembling...

asl -xx -c $debug_syms $print_err $listing_opt $multipass_warn -A -U s2.asm

if [[ -f s2.log ]]; then
	echo
	echo "*****************************************"
	echo "*                                       *"
	echo "*   There were build errors/warnings.   *"
	echo "*                                       *"
	echo "*****************************************"
	echo
	cat s2.log
	exit 1
fi

[[ -f s2.p ]] && bin/s2p2bin_plus s2.p KEH.bin s2.h
[[ -f KEH.bin ]] && bin/fixpointer s2.h KEH.bin
[[ -f KEH.bin ]] && bin/fixheader KEH.bin

