﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace KEHcredits
{
	class Program
	{
		static readonly Dictionary<string, string> levelnames = new Dictionary<string, string>()
		{
			{ "EHZ", "Emerald Hill Zone:" },
			{ "CPZ", "Chemical Plant Zone:" },
			{ "ARZ", "Aquatic Ruin Zone:" },
			{ "CNZ", "Casino Night Zone:" },
			{ "HTZ", "Hill Top Zone:" },
			{ "MCZ", "Mystic Cave Zone:" },
			{ "OOZ", "Oil Ocean Zone:" },
			{ "MTZ", "Metropolis Zone:" },
			{ "WFZ", "Wing Fortress Zone:" },
			{ "MZ", "Marble Zone:" },
			{ "SYZ", "Spring Yard Zone:" },
			{ "LZ", "Labyrinth Zone:" },
			{ "HPZ", "Hidden Palace Zone:" }
		};

		static void Main(string[] args)
		{
			List<string> lines = new List<string>(File.ReadAllLines("credits.txt"));
			int highscoreline = lines.IndexOf("<highscores>");
			lines.RemoveAt(highscoreline);
			List<int> dups = new List<int>(lines.Count);
			dups.Add(0);
			for (int i = 1; i < lines.Count; i++)
			{
				dups.Add(i);
				for (int j = 0; j < i; j++)
					if (lines[i].Equals(lines[j], StringComparison.OrdinalIgnoreCase))
					{
						dups[i] = j;
						break;
					}
			}
			using (StreamWriter writer = File.CreateText("credits.asm"))
			{
				writer.WriteLine("off_B2CA:");
				int i = 0;
				foreach (string line in lines)
				{
					if (!string.IsNullOrWhiteSpace(line))
					{
						int len = 0;
						foreach (char c in line)
							switch (c)
							{
								default:
									len += 2;
									break;
								case ' ':
								case 'I':
								case 'i':
								case '.':
								case '*':
									len++;
									break;
								case ':':
									break;
							}
						writer.WriteLine("\tcreditsPtr\tcred_line{0}, {1}, {2}", dups[i] + 1,
							20 - (len / 2), i * 2);
					}
					++i;
				}
				writer.WriteLine("\tdc.l\t-1");
				writer.WriteLine();
				i = 0;
				foreach (string line in lines)
				{
					if (dups[i] == i && !string.IsNullOrWhiteSpace(line))
						writer.WriteLine("cred_line{0}:\tcreditText\t{1},\"{2}\"", i + 1, line.EndsWith(":") ? 1 : 0,
							line.TrimEnd(':').ToUpperInvariant());
					++i;
				}
				writer.WriteLine();
				List<string> newlines = new List<string>();
				List<int> newdups = new List<int>();
				newlines.Add("Tester High Scores:");
				newdups.Add(0);
				newlines.Add(string.Empty);
				newdups.Add(1);
				using (Stream str = File.OpenRead("keh.csv"))
				using (StreamReader sr = new StreamReader(str))
				{
					string line;
					while (!string.IsNullOrEmpty(line = sr.ReadLine()))
					{
						string[] split = line.Split(',');
						newlines.Add(levelnames[split[0]]);
						newdups.Add(newdups.Count);
						newlines.Add(split[2]);
						newdups.Add(newdups.Count);
						for (int j = 0; j < lines.Count; j++)
							if (split[2].Equals(lines[j], StringComparison.OrdinalIgnoreCase))
							{
								newdups[newdups.Count - 1] = -j;
								break;
							}
						if (newdups[newdups.Count - 1] > 0)
							for (int j = 0; j < newlines.Count - 1; j++)
								if (split[2].Equals(newlines[j], StringComparison.OrdinalIgnoreCase))
								{
									newdups[newdups.Count - 1] = j;
									break;
								}
						newlines.Add(split[1]);
						newdups.Add(newdups.Count);
						newlines.Add(string.Empty);
						newdups.Add(newdups.Count);
					}
				}
				for (int j = highscoreline; j < lines.Count; j++)
				{
					newlines.Add(lines[j]);
					newdups.Add(-dups[j]);
				}
				writer.WriteLine("off_B2CA_2:");
				i = 0;
				foreach (string line in lines)
				{
					if (!string.IsNullOrWhiteSpace(line))
					{
						int len = 0;
						foreach (char c in line)
							switch (c)
							{
								default:
									len += 2;
									break;
								case ' ':
								case 'I':
								case 'i':
								case '.':
								case '*':
									len++;
									break;
								case ':':
									break;
							}
						writer.WriteLine("\tcreditsPtr\tcred_line{0}, {1}, {2}", dups[i] + 1,
							20 - (len / 2), i * 2);
					}
					if (++i == highscoreline)
						break;
				}
				foreach (string line in newlines)
				{
					if (!string.IsNullOrWhiteSpace(line))
					{
						int len = 0;
						foreach (char c in line)
							switch (c)
							{
								default:
									len += 2;
									break;
								case ' ':
								case 'I':
								case 'i':
								case '.':
								case '*':
									len++;
									break;
								case ':':
									break;
							}
						if (newdups[i - highscoreline] < 0)
							writer.WriteLine("\tcreditsPtr\tcred_line{0}, {1}, {2}", -newdups[i - highscoreline] + 1,
								20 - (len / 2), i * 2);
						else
							writer.WriteLine("\tcreditsPtr\tcred2_line{0}, {1}, {2}", newdups[i - highscoreline] + 1,
								20 - (len / 2), i * 2);
					}
					++i;
				}
				writer.WriteLine("\tdc.l\t-1");
				writer.WriteLine();
				i = 0;
				foreach (string line in newlines)
				{
					if (newdups[i] == i && !string.IsNullOrWhiteSpace(line))
						writer.WriteLine("cred2_line{0}:\tcreditText\t{1},\"{2}\"", i + 1, line.EndsWith(":") ? 1 : 0,
							line.TrimEnd(':').ToUpperInvariant());
					++i;
				}
				writer.Flush();
				writer.Close();
			}
		}
	}
}