﻿using IniFile;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;

namespace KEHmusic
{
	static class Program
	{
		static Dictionary<string, string> PlaylistNames = new Dictionary<string, string>()
		{
			{"Title", "Title Screen"},
			{"Menu", "Menus"},
			{"EHZ", "Emerald Hill Zone"},
			{"CPZ", "Chemical Plant Zone"},
			{"ARZ", "Aquatic Ruin Zone"},
			{"CNZ", "Casino Night Zone"},
			{"HTZ", "Hill Top Zone"},
			{"MCZ", "Mystic Cave Zone"},
			{"OOZ", "Oil Ocean Zone"},
			{"MTZ", "Metropolis Zone"},
			{"WFZ", "Wing Fortress Zone"},
			{"MZ", "Marble Zone"},
			{"SYZ", "Spring Yard Zone"},
			{"LZ", "Labyrinth Zone"},
			{"SBZ", "Labyrinth Zone (SBZ Palette)"},
			{"HPZ", "Hidden Palace Zone"},
			{"EHZ_2P", "Emerald Hill Zone 2P"},
			{"CNZ_2P", "Casino Night Zone 2P"},
			{"MCZ_2P", "Mystic Cave Zone 2P"},
			{"Invincibility", "Invincibility"},
			{"ActClear", "Zone Clear"},
			{"GameOver", "Game Over"},
			{"ExtraLife", "Extra Life"},
			{"Credits", "Credits"}
		};
		static void Main(string[] args)
		{
			var dacs = IniSerializer.Deserialize<Dictionary<string, DACInfo>>("sound/DAC/DAC.ini");
			using (StreamWriter sw = new StreamWriter("sound/SMPSPlay/DAC.ini", false, Encoding.ASCII))
			{
				int baseCycles = 227;
				int loopCycles = 26;
				int loopSamples = 2;
				double baseRate = 3579545.0 * loopSamples / loopCycles;
				double rateDiv = (double)baseCycles / loopCycles - 1.0;
				sw.WriteLine("; Knuckles Emerald Hunt DAC driver rates");
				sw.WriteLine("BaseRate = {0}", baseRate.ToString("F0", CultureInfo.InvariantCulture));
				sw.WriteLine("RateDiv = {0}", rateDiv.ToString("F2", CultureInfo.InvariantCulture));
				sw.WriteLine("");
				sw.WriteLine("BaseCycles = {0}", baseCycles);
				sw.WriteLine("LoopCycles = {0}", loopCycles);
				sw.WriteLine("LoopSamples = {0}", loopSamples);
				sw.WriteLine("");
				
				int dacID = 0x81;
				foreach (var item in dacs)
				{
					sw.WriteLine("[{0:X2}]\t; {1}", dacID, item.Key);
					sw.WriteLine("Compr = DPCM");
					sw.WriteLine("File = ../DAC/{0}", item.Value.File);
					sw.WriteLine("Rate = 0x{0:X2}", item.Value.Delay);
					sw.WriteLine("");
					dacID++;
				}
			}
			
			var songs = IniSerializer.Deserialize<Dictionary<string, MusicInfo>>("sound/music/music.ini");
			var songplist = new Dictionary<string, List<Tuple<string, int>>>();
			foreach (var item in songs)
				songplist[item.Key] = new List<Tuple<string, int>>();
			
			var playlistini = IniFile.IniFile.Load("musiclist.ini");
			var playlist = new Dictionary<string, string[]>(playlistini.Count);
			foreach (var item in playlistini[string.Empty])
			{
				playlist[item.Key] = item.Value.Split(new[] { ", " }, StringSplitOptions.None);
				for (int songplidx = 0; songplidx < playlist[item.Key].Length; songplidx++)
					songplist[playlist[item.Key][songplidx]].Add(new Tuple<string,int>(item.Key, songplidx));
			}
			
			List<string> vgmTagList = new List<string>();
			List<string> vgmNotesList = new List<string>();
			int songID = 0x01;
			foreach (var item in songs)
			{
				string filePrefix = songID.ToString("X2") + " ";
				string smpsFileName = "";
				switch (Path.GetExtension(item.Value.File).ToLowerInvariant())
				{
				case ".asm":
				case ".68k":
					{
						bool ramOffsets = (item.Value.Compressible && item.Value.Size <= 0x800);
						int songOfs = ramOffsets ? 0x1380 : 0x8000;
						using (StreamWriter sw = new StreamWriter("tmp.asm", false, Encoding.ASCII))
						{
							sw.WriteLine("\tCPU 68000");
							sw.WriteLine("\tpadding off");
							if ((songOfs & 0x8000) == 0x0000)
								sw.WriteLine("kehmusic = 1");
							sw.WriteLine("allOptimizations = 1");
							sw.WriteLine("\tinclude \"s2.macros.asm\"");
							sw.WriteLine("\tinclude \"sound/_smps2asm_inc.asm\"");
							sw.WriteLine("\tphase ${0:X4}", songOfs);
							sw.WriteLine("\tinclude \"sound/music/{0}\"", item.Value.File);
						}
						
						ProcessStartInfo si = new ProcessStartInfo("win32/asw", "-E -xx -A -r 2 -q -U tmp.asm");
						si.CreateNoWindow = true;
						si.EnvironmentVariables.Add("AS_MSGPATH", "win32/msg");
						si.EnvironmentVariables.Add("USEANSI", "n");
						si.UseShellExecute = false;
						using (Process proc = Process.Start(si))
							proc.WaitForExit();
						File.Delete("tmp.asm");
						if (File.Exists("tmp.log"))
						{
							Console.Write(File.ReadAllText("tmp.log"));
							File.Delete("tmp.log");
						}
						if (!File.Exists("tmp.p"))
							continue;
						
						si = new ProcessStartInfo("win32/s2p2bin_plus", "tmp.p tmp.bin");
						si.CreateNoWindow = true;
						si.UseShellExecute = false;
						using (Process proc = Process.Start(si))
							proc.WaitForExit();
						File.Delete("tmp.p");
						if (!File.Exists("tmp.bin"))
							continue;
						
						string destPath;
						smpsFileName = Path.GetFileNameWithoutExtension(item.Value.File);
						// add Z80 offset to file name for songs where SMPSPlay's autodetection fails
						if (item.Value.OfsInName)
							smpsFileName += "." + songOfs.ToString("X4");
						smpsFileName = filePrefix + smpsFileName + ".sm2";
						destPath = Path.Combine("sound/SMPSPlay", smpsFileName);
						if (File.Exists(destPath))
							File.Delete(destPath);
						File.Move("tmp.bin", destPath);
					}
					break;
				case ".bin":
					smpsFileName = filePrefix + item.Value.File;
					File.Copy(Path.Combine("sound/music", item.Value.File), Path.Combine("sound/SMPSPlay", smpsFileName));
					break;
				}
				string vgmTagParams;
				string titleTrimmed;
				string authorTrimmed;
				List<string> notesTags;
				
				titleTrimmed = (item.Value.Title == null) ? "" : item.Value.Title.Trim();
				authorTrimmed = (item.Value.Author == null) ? "" : item.Value.Author.Trim();
				if (authorTrimmed.Length == 0)
					authorTrimmed = "Sonic Team";
				notesTags = new List<string>();
				foreach (var plitem in songplist[item.Key])
				{
					int choiceidx = 1 + plitem.Item2;
					if (plitem.Item1 == "SBZ")
						choiceidx += playlist["LZ"].Length;
					notesTags.Add(String.Format("Song choice #{0} for {1}", choiceidx, PlaylistNames[plitem.Item1]));
				}
				
				vgmTagParams = String.Format("-Title:\"{0}\" -Author:\"{1}\" -Notes:\"{2}\" \"{3}\"",
											titleTrimmed, authorTrimmed, String.Join("\\n", notesTags),
											Path.GetFileNameWithoutExtension(smpsFileName) + ".vgm");
				vgmTagList.Add(vgmTagParams);
				songID++;
			}
			using (StreamWriter sw = new StreamWriter("sound/SMPSPlay/tagvgms.bat", false, Encoding.ASCII))
			{
				string vgmtag_tool = "vgm_tag.exe";
				foreach (var item in vgmTagList)
					sw.WriteLine(vgmtag_tool + " " + item);
				sw.WriteLine("pause");
			}
		}
	}

	class MusicInfo
	{
		public string Title { get; set; }
		public string Author { get; set; }
		public string File { get; set; }
		public bool PALMode { get; set; }
		[DefaultValue(true)]
		public bool Compressible { get; set; }
		public bool OfsInName { get; set; }
		public short Size { get; set; }
		public string MD5 { get; set; }
		public string OutputFile { get; set; }
	}

	class DACInfo
	{
		public string File { get; set; }
		[TypeConverter(typeof(ByteHexConverter))]
		public byte Delay { get; set; }
	}

	public class ByteHexConverter : TypeConverter
	{
		public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
		{
			if (destinationType == typeof(string))
				return true;
			return base.CanConvertTo(context, destinationType);
		}

		public override object ConvertTo(ITypeDescriptorContext context, CultureInfo culture, object value, Type destinationType)
		{
			if (destinationType == typeof(string) && value is byte)
				return ((byte)value).ToString("X");
			return base.ConvertTo(context, culture, value, destinationType);
		}

		public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
		{
			if (sourceType == typeof(string))
				return true;
			return base.CanConvertFrom(context, sourceType);
		}

		public override object ConvertFrom(ITypeDescriptorContext context, CultureInfo culture, object value)
		{
			if (value is string)
				return byte.Parse((string)value, NumberStyles.HexNumber);
			return base.ConvertFrom(context, culture, value);
		}

		public override bool IsValid(ITypeDescriptorContext context, object value)
		{
			if (value is byte)
				return true;
			byte i;
			if (value is string)
				return byte.TryParse((string)value, NumberStyles.HexNumber, NumberFormatInfo.InvariantInfo, out i);
			return base.IsValid(context, value);
		}
	}
}
