; >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
; Equates section - Names for variables.

	dottedstructs on

; ---------------------------------------------------------------------------
; size variables - you'll get an informational error if you need to change these...
; they are all in units of bytes
Size_of_Snd_driver_guess =	$E14 ; approximate post-compressed size of the Z80 sound driver

; ---------------------------------------------------------------------------
; Object Status Table offsets (for everything between Object_RAM and Primary_Collision)
; ---------------------------------------------------------------------------
; universally followed object conventions:
id =			  0 ; object ID (if you change this, change insn1op and insn2op in s2.macrosetup.asm, if you still use them)
render_flags =		  1 ; bitfield ; bit 7 = onscreen flag, bit 0 = x mirror, bit 1 = y mirror, bit 2 = coordinate system, bit 6 = render subobjects
art_tile =		  2 ; and 3 ; start of sprite's art
mappings =		  4 ; and 5 and 6 and 7
x_pos =			  8 ; and 9 ... some objects use $A and $B as well when extra precision is required (see ObjectMove) ... for screen-space objects this is called x_pixel instead
x_sub =			 $A ; and $B
y_pos =			 $C ; and $D ... some objects use $E and $F as well when extra precision is required ... screen-space objects use y_pixel instead
y_sub =			 $E ; and $F
priority =		$18 ; 0 = front
width_pixels =		$19
mapping_frame =		$1A
; ---------------------------------------------------------------------------
; conventions followed by most objects:
x_vel =			$10 ; and $11 ; horizontal velocity
y_vel =			$12 ; and $13 ; vertical velocity
y_radius =		$16 ; collision height / 2
x_radius =		$17 ; collision width / 2
anim_frame =		$1B
anim =			$1C
next_anim =		$1D
anim_frame_duration =	$1E
anim_delay =	$1F
status =		$22 ; note: exact meaning depends on the object... for sonic/tails: bit 0: leftfacing. bit 1: inair. bit 2: spinning. bit 3: onobject. bit 4: rolljumping. bit 5: pushing. bit 6: underwater.
routine =		$24
routine_secondary =	$25
angle =			$26 ; angle about the z axis (360 degrees = 256)
; ---------------------------------------------------------------------------
; conventions followed by many objects but NOT sonic/tails:
collision_flags =	$20
collision_property =	$21
respawn_index =		$23
subtype =		$28
; ---------------------------------------------------------------------------
; conventions specific to sonic/tails (Obj01, Obj02, and ObjDB):
; note: $1F, $20, and $21 are unused and available
inertia =		$14 ; and $15 ; directionless representation of speed... not updated in the air
glidemode =		$21
glideflags =	$20
glideunk =		$1F
flip_angle =		$27 ; angle about the x axis (360 degrees = 256) (twist/tumble)
air_left =		$28
flip_turned =		$29 ; 0 for normal, 1 to invert flipping (it's a 180 degree rotation about the axis of Sonic's spine, so he stays in the same position but looks turned around)
obj_control =		$2A ; 0 for normal, 1 for hanging or for resting on a flipper, $81 for going through CNZ/OOZ/MTZ tubes or stopped in CNZ cages or stoppers or flying if Tails
status_secondary =	$2B
flips_remaining =	$2C ; number of flip revolutions remaining
flip_speed =		$2D ; number of flip revolutions per frame / 256
move_lock =		$2E ; and $2F ; horizontal control lock, counts down to 0
invulnerable_time =	$30 ; and $31 ; time remaining until you stop blinking
invincibility_time =	$32 ; and $33 ; remaining
speedshoes_time =	$34 ; and $35 ; remaining
next_tilt =		$36 ; angle on ground in front of sprite
tilt =			$37 ; angle on ground
stick_to_convex =	$38 ; 0 for normal, 1 to make Sonic stick to convex surfaces like the rotating discs in Sonic 1 and 3 (unused in Sonic 2 but fully functional)
spindash_flag =		$39 ; 0 for normal, 1 for charging a spindash or forced rolling
spindash_counter =	$3A ; and $3B
jumping =		$3C
interact =		$3D ; RAM address of the last object Sonic stood on, minus $FFFFB000 and divided by $40
layer =			$3E ; collision plane, track switching...
layer_plus =		$3F ; always same as layer+1 ?? used for collision somehow
; ---------------------------------------------------------------------------
; conventions followed by several objects but NOT sonic/tails:
y_pixel =		2+x_pos ; and 3+x_pos ; y coordinate for objects using screen-space coordinate system
x_pixel =		x_pos ; and 1+x_pos ; x coordinate for objects using screen-space coordinate system
parent =		$3E ; and $3F ; address of object that owns or spawned this one, if applicable
; TODO: $2C is often parent instead (see LoadChildObject); consider defining parent2 = $2C and changing some objoff_2Cs to that
; ---------------------------------------------------------------------------
; conventions followed by some/most bosses:
boss_subtype		= $A
boss_invulnerable_time	= $14
boss_sine_count		= $1A	;mapping_frame
boss_routine		= $26	;angle
boss_defeated		= $2C
boss_hitcount2		= $32
boss_hurt_sonic		= $38	; flag set by collision response routine when sonic has just been hurt (by boss?)
; ---------------------------------------------------------------------------
; when childsprites are activated (i.e. bit #6 of render_flags set)
mainspr_mapframe	= $B
mainspr_width		= $E
mainspr_childsprites 	= $F	; amount of child sprites
mainspr_height		= $14
sub2_x_pos		= $10	;x_vel
sub2_y_pos		= $12	;y_vel
sub2_mapframe		= $15
sub3_x_pos		= $16	;y_radius
sub3_y_pos		= $18	;priority
sub3_mapframe		= $1B	;anim_frame
sub4_x_pos		= $1C	;anim
sub4_y_pos		= $1E	;anim_frame_duration
sub4_mapframe		= $21	;collision_property
sub5_x_pos		= $22	;status
sub5_y_pos		= $24	;routine
sub5_mapframe		= $27
sub6_x_pos		= $28	;subtype
sub6_y_pos		= $2A
sub6_mapframe		= $2D
sub7_x_pos		= $2E
sub7_y_pos		= $30
sub7_mapframe		= $33
sub8_x_pos		= $34
sub8_y_pos		= $36
sub8_mapframe		= $39
sub9_x_pos		= $3A
sub9_y_pos		= $3C
sub9_mapframe		= $3F
next_subspr		= $6
; ---------------------------------------------------------------------------
; unknown or inconsistently used offsets that are not applicable to sonic/tails:
; (provided because rearrangement of the above values sometimes requires making space in here too)
objoff_A =		2+x_pos ; note: x_pos can be 4 bytes, but sometimes the last 2 bytes of x_pos are used for other unrelated things
objoff_B =		3+x_pos ; unused
objoff_E =		2+y_pos	; unused
objoff_F =		3+y_pos ; unused
objoff_10 =		$10
objoff_14 =		$14
objoff_15 =		$15
objoff_1F =		$1F
objoff_27 =		$27
objoff_28 =		$28 ; overlaps subtype, but a few objects use it for other things anyway
 enum               objoff_29=$29,objoff_2A=$2A,objoff_2B=$2B,objoff_2C=$2C,objoff_2D=$2D,objoff_2E=$2E,objoff_2F=$2F
 enum objoff_30=$30,objoff_31=$31,objoff_32=$32,objoff_33=$33,objoff_34=$34,objoff_35=$35,objoff_36=$36,objoff_37=$37
 enum objoff_38=$38,objoff_39=$39,objoff_3A=$3A,objoff_3B=$3B,objoff_3C=$3C,objoff_3D=$3D,objoff_3E=$3E,objoff_3F=$3F
; ---------------------------------------------------------------------------
; property of all objects:
object_size =		$40 ; the size of an object
next_object =		object_size

RadarFrame	STRUCT
cur	ds.b	1
prev	ds.b	1
	ENDSTRUCT

EmeraldInfo	STRUCT
rom	ds.l	1
ram	ds.w	1
	ENDSTRUCT

EmeraldManager	STRUCT
id	ds.b	1
	ds.b	$D
radar1	RadarFrame
radar2	RadarFrame
radar3	RadarFrame
radar4	RadarFrame
radar5	RadarFrame
emerald1	EmeraldInfo
emerald2	EmeraldInfo
emerald3	EmeraldInfo
emerald4	EmeraldInfo
emerald5	EmeraldInfo
p1emecnt	ds.b	1
p2emecnt	ds.b	1
pingtime	ds.w	1
emeraldframes	ds.w	1
bonusframe	ds.b	1
bonustime	ds.b	1
emeraldint	ds.b	1
gotemeralds	ds.b	1
	ENDSTRUCT

	if EmeraldManager.len <> object_size
		fatal "EmeraldManager structure is $\{EmeraldManager.len} bytes, but should be $\{object_size}."
	endif

; ---------------------------------------------------------------------------
; Bits 3-6 of an object's status after a SolidObject call is a
; bitfield with the following meaning:
p1_standing_bit   = 3
p2_standing_bit   = p1_standing_bit + 1

p1_standing       = 1<<p1_standing_bit
p2_standing       = 1<<p2_standing_bit

pushing_bit_delta = 2
p1_pushing_bit    = p1_standing_bit + pushing_bit_delta
p2_pushing_bit    = p1_pushing_bit + 1

p1_pushing        = 1<<p1_pushing_bit
p2_pushing        = 1<<p2_pushing_bit


standing_mask     = p1_standing|p2_standing
pushing_mask      = p1_pushing|p2_pushing

; ---------------------------------------------------------------------------
; The high word of d6 after a SolidObject call is a bitfield
; with the following meaning:
p1_touch_side_bit   = 0
p2_touch_side_bit   = p1_touch_side_bit + 1

p1_touch_side       = 1<<p1_touch_side_bit
p2_touch_side       = 1<<p2_touch_side_bit

touch_side_mask     = p1_touch_side|p2_touch_side

p1_touch_bottom_bit = p1_touch_side_bit + pushing_bit_delta
p2_touch_bottom_bit = p1_touch_bottom_bit + 1

p1_touch_bottom     = 1<<p1_touch_bottom_bit
p2_touch_bottom     = 1<<p2_touch_bottom_bit

touch_bottom_mask   = p1_touch_bottom|p2_touch_bottom

p1_touch_top_bit   = p1_touch_bottom_bit + pushing_bit_delta
p2_touch_top_bit   = p1_touch_top_bit + 1

p1_touch_top       = 1<<p1_touch_top_bit
p2_touch_top       = 1<<p2_touch_top_bit

touch_top_mask     = p1_touch_top|p2_touch_top

; ---------------------------------------------------------------------------
; Controller Buttons
;
; Buttons bit numbers
button_up:			EQU	0
button_down:			EQU	1
button_left:			EQU	2
button_right:			EQU	3
button_B:			EQU	4
button_C:			EQU	5
button_A:			EQU	6
button_start:			EQU	7
; Buttons masks (1 << x == pow(2, x))
button_up_mask:			EQU	1<<button_up	; $01
button_down_mask:		EQU	1<<button_down	; $02
button_left_mask:		EQU	1<<button_left	; $04
button_right_mask:		EQU	1<<button_right	; $08
button_B_mask:			EQU	1<<button_B	; $10
button_C_mask:			EQU	1<<button_C	; $20
button_A_mask:			EQU	1<<button_A	; $40
button_start_mask:		EQU	1<<button_start	; $80

; ---------------------------------------------------------------------------
; Casino night bumpers
bumper_id           = 0
bumper_x            = 2
bumper_y            = 4
next_bumper         = 6
prev_bumper_x       = bumper_x-next_bumper

; ---------------------------------------------------------------------------
; Constants that can be used instead of hard-coded IDs for various things.
; The "id" function allows to remove elements from an array/table without having
; to change the IDs everywhere in the code.

cur_zone_id := 0 ; the zone ID currently being declared
cur_zone_str := "0" ; string representation of the above

; macro to declare a zone ID
; this macro also declares constants of the form zone_id_X, where X is the ID of the zone in stock Sonic 2
; in order to allow level offset tables to be made dynamic
zoneID macro zoneID,{INTLABEL}
__LABEL__ = zoneID
zone_id_{cur_zone_str} = zoneID
cur_zone_id := cur_zone_id+1
cur_zone_str := "\{cur_zone_id}"
    endm

; Zone IDs. These MUST be declared in the order in which their IDs are in stock Sonic 2, otherwise zone offset tables will screw up
emerald_hill_zone zoneID	$00
chemical_plant_zone zoneID			$01
aquatic_ruin_zone zoneID		$02
casino_night_zone zoneID			$03
hill_top_zone zoneID			$04
mystic_cave_zone zoneID	$05
oil_ocean_zone zoneID	$06
metropolis_zone zoneID		$07
wing_fortress_zone zoneID	$08
marble_zone zoneID		$09
spring_yard_zone zoneID		$0A
labyrinth_zone zoneID	$0B
hidden_palace_zone zoneID	$0C
lattice_zone zoneID	$0D

; NOTE: If you want to shift IDs around, set useFullWaterTables to 1 in the assembly options

; V-Int routines

; some variables to help define those constants (redefined before a new set of IDs)
offset :=	Vint_SwitchTbl	; this is the start of the pointer table
ptrsize :=	1		; this is the size of a pointer (should be 1 if the ID is a multiple of the actual size)
idstart :=	0		; value to add to all IDs

; function using these variables
id function ptr,((ptr-offset)/ptrsize+idstart)

VintID_Lag =		id(Vint_Lag_ptr) ; 0
VintID_SEGA =		id(Vint_SEGA_ptr) ; 2
VintID_Title =		id(Vint_Title_ptr) ; 4
VintID_Level =		id(Vint_Level_ptr) ; 8
VintID_TitleCard =	id(Vint_TitleCard_ptr) ; C
VintID_Pause =		id(Vint_Level_ptr) ; 10
VintID_Fade =		id(Vint_Fade_ptr) ; 12
VintID_PCM =		id(Vint_PCM_ptr) ; 14
VintID_Menu =		id(Vint_Menu_ptr) ; 16
VintID_Ending =		id(Vint_Ending_ptr) ; 18
VintID_CtrlDMA =	id(Vint_CtrlDMA_ptr) ; 1A
VintID_LevelSelect =	id(Vint_LevelSelect_ptr)

; set the number of zones
no_of_zones = cur_zone_id


; Game modes
offset :=	GameModesArray
ptrsize :=	4
idstart :=	0

gameModeID	macro	label
GameModeID_label =	id(GameMode_label)
	endm

	gameModeID SegaScreen ; 0
	gameModeID TitleScreen ; 4
	gameModeID Demo ; 8
	gameModeID Level ; C
	gameModeID 2PResults ; 18
	gameModeID 2PLevelSelect ; 1C
	gameModeID EndingSequence ; 20
	gameModeID OptionsMenu ; 24
	gameModeID LevelSelect ; 28
	gameModeID MusicSwitch ; 2C
GameModeFlag_TitleCard =	7 ; flag bit
GameModeID_TitleCard =		1<<GameModeFlag_TitleCard ; flag mask

makeID	macro	label
{idtype}ID_label =	id({idtype}Ptr_label)
	endm

; palette IDs
offset :=	PalPointers
ptrsize :=	8
idstart :=	0
idtype :=	"Pal"

	makeID SEGA ; 0
	makeID Title ; 1
	makeID UNK1 ; 2
	makeID BGND ; 3
	makeID EHZ ; 4
	makeID MTZ ; 8
	makeID WFZ ; A
	makeID HTZ ; B
	makeID HPZ ; C
	makeID OOZ ; E
	makeID MCZ ; F
	makeID CNZ ; 10
	makeID CPZ ; 11
	makeID ARZ ; 13
	makeID HPZ_U ; 15
	makeID CPZ_U ; 16
	makeID ARZ_U ; 17
	makeID UNK2 ; 19
	makeID UNK3 ; 1A
	makeID Menu ; 26
	makeID MZ ; 28
	makeID SYZ ; 28
	makeID LZ ; 28
	makeID SBZ ; 28
	makeID LZ_U ; 28
	makeID SBZ_U ; 28
	makeID Congratulations
	makeID Lattice

; PLC IDs
offset :=	ArtLoadCues
ptrsize :=	2
idstart :=	0
idtype :=	"PLC"

	makeID Std1 ; 0
	makeID Std2 ; 1
	makeID StdWtr ; 2
	makeID StdS1 ; 2
	makeID StdWfz ; 2
	makeID GameOver ; 3
	makeID Ehz1 ; 4
	makeID Ehz2 ; 5
	makeID Tails1up ; 8
	makeID TailsLife ; 9
	makeID Mtz1 ; C
	makeID Mtz2 ; D
	makeID Wfz1 ; 10
	makeID Wfz2 ; 11
	makeID Htz1 ; 12
	makeID Htz2 ; 13
	makeID Hpz1 ; 14
	makeID Hpz2 ; 15
	makeID Ooz1 ; 18
	makeID Ooz2 ; 19
	makeID Mcz1 ; 1A
	makeID Mcz2 ; 1B
	makeID Cnz1 ; 1C
	makeID Cnz2 ; 1D
	makeID Cpz1 ; 1E
	makeID Cpz2 ; 1F
	makeID Arz1 ; 22
	makeID Arz2 ; 23
	makeID Mz1
	makeID Mz2
	makeID Syz1
	makeID Syz2
	makeID Lz1
	makeID Lz2
	makeID Results ; 26
	makeID FieryExplosion ; 30
	makeID EhzAnimals ; 32
	makeID MczAnimals ; 33
	makeID HtzAnimals ; 34
	makeID MtzAnimals ; 34
	makeID WfzAnimals ; 34
	makeID HpzAnimals ; 36
	makeID OozAnimals ; 37
	makeID SczAnimals ; 38
	makeID CnzAnimals ; 39
	makeID CpzAnimals ; 3A
	makeID ArzAnimals ; 3B
	makeID Tornado ; 3F
	makeID Capsule ; 40
	makeID Explosion ; 41
	makeID Results2P
	makeID MzAnimals ; 4A
	makeID SyzAnimals ; 4E
	makeID LzAnimals ; 52

; Object IDs
offset :=	Obj_Index
ptrsize :=	4
idstart :=	1
idtype :=	"Obj"

	makeID Sonic		; 01
	makeID Tails		; 02
	makeID PlaneSwitcher	; 03
	makeID WaterSurface		; 04
	makeID TailsTails		; 05
	makeID Spiral		; 06
	makeID Oil			; 07
	makeID SpindashDust		; 08
	makeID Splash		; 08
	makeID SmallBubbles		; 0A
	makeID TippingFloor		; 0B
	makeID Signpost		; 0D
	makeID Logo		; 0E
	makeID TitleMenu		; 0F
	makeID Bridge		; 11
	makeID HPZEmerald		; 12
	makeID HPZWaterfall		; 13
	makeID Seesaw		; 14
	makeID SwingingPlatform	; 15
	makeID HTZLift		; 16
	makeID ARZPlatform		; 18
	makeID EHZPlatform		; 18
	makeID CPZPlatform		; 19
	makeID OOZMovingPform	; 19
	makeID WFZPlatform		; 19
	makeID HPZCollapsPform	; 1A
	makeID SpeedBooster		; 1B
	makeID Scenery		; 1C
	makeID BridgeStake		; 1C
	makeID FallingOil		; 1C
	makeID BlueBalls		; 1D
	makeID CPZSpinTube		; 1E
	makeID CollapsPform		; 1F
	makeID LavaBubble		; 20
	makeID HUD			; 21
	makeID ArrowShooter		; 22
	makeID FallingPillar	; 23
	makeID ARZBubbles		; 24
	makeID Ring			; 25
	makeID Monitor		; 26
	makeID Explosion		; 27
	makeID Animal		; 28
	makeID Points		; 29
	makeID Stomper		; 2A
	makeID RisingPillar		; 2B
	makeID LeavesGenerator	; 2C
	makeID Barrier		; 2D
	makeID MonitorContents	; 2E
	makeID SmashableGround	; 2F
	makeID RisingLava		; 30
	makeID LavaMarker		; 31
	makeID BreakableBlock	; 32
	makeID BreakableRock	; 32
	makeID OOZPoppingPform	; 33
	makeID TitleCard		; 34
	makeID InvStars		; 35
	makeID Spikes		; 36
	makeID LostRings		; 37
	makeID Shield		; 38
	makeID GameOver		; 39
	makeID TimeOver		; 39
	makeID Results		; 3A
	makeID SolidBlock		; 3B
	makeID OOZLauncher		; 3D
	makeID EggPrison		; 3E
	makeID Fan			; 3F
	makeID Springboard		; 40
	makeID Spring		; 41
	makeID SteamSpring		; 42
	makeID SlidingSpike		; 43
	makeID RoundBumper		; 44
	makeID OOZSpring		; 45
	makeID OOZBall		; 46
	makeID Button		; 47
	makeID LauncherBall		; 48
	makeID EHZWaterfall		; 49
	makeID Octus		; 4A
	makeID Buzzer		; 4B
	makeID EmeraldManager
	makeID Batbot
	makeID Emerald
	makeID Dinobot
	makeID Aquis		; 50
	makeID CNZBoss		; 51
	makeID HTZBoss		; 52
	makeID MTZBossOrb		; 53
	makeID MTZBoss		; 54
	makeID OOZBoss		; 55
	makeID EHZBoss		; 56
	makeID MCZBoss		; 57
	makeID BossExplosion	; 58
	makeID Masher		; 5C
	makeID CPZBoss		; 5D
	makeID MTZTwinStompers	; 64
	makeID MTZLongPlatform	; 65
	makeID MTZSpringWall	; 66
	makeID MTZSpinTube		; 67
	makeID SpikyBlock		; 68
	makeID Nut			; 69
	makeID MCZRotPforms		; 6A
	makeID MTZMovingPforms	; 6A
	makeID MTZPlatform		; 6B
	makeID CPZSquarePform	; 6B
	makeID Conveyor		; 6C
	makeID FloorSpike		; 6D
	makeID LargeRotPform	; 6E
	makeID Cog			; 70
	makeID MTZLavaBubble	; 71
	makeID HPZBridgeStake	; 71
	makeID PulsingOrb		; 71
	makeID CNZConveyorBelt	; 72
	makeID RotatingRings	; 73
	makeID InvisibleBlock	; 74
	makeID MCZBrick		; 75
	makeID SlidingSpikes	; 76
	makeID MCZBridge		; 77
	makeID CPZStaircase		; 78
	makeID Starpost		; 79
	makeID SidewaysPform	; 7A
	makeID PipeExitSpring	; 7B
	makeID CPZPylon		; 7C
	makeID SuperSonicStars	; 7E
	makeID VineSwitch		; 7F
	makeID MovingVine		; 80
	makeID MCZDrawbridge	; 81
	makeID SwingingPform	; 82
	makeID ARZRotPforms		; 83
	makeID ForcedSpin		; 84
	makeID PinballMode		; 84
	makeID LauncherSpring	; 85
	makeID Flipper		; 86
	makeID ARZBoss		; 89
	makeID WFZPalSwitcher	; 8B
	makeID Whisp		; 8C
	makeID GrounderInWall	; 8D
	makeID GrounderInWall2	; 8E
	makeID GrounderWall		; 8F
	makeID GrounderRocks	; 90
	makeID ChopChop		; 91
	makeID Spiker		; 92
	makeID SpikerDrill		; 93
	makeID Rexon		; 94
	makeID Sol			; 95
	makeID Rexon2		; 96
	makeID RexonHead		; 97
	makeID Projectile		; 98
	makeID Nebula		; 99
	makeID Turtloid		; 9A
	makeID TurtloidRider	; 9B
	makeID BalkiryJet		; 9C
	makeID Coconuts		; 9D
	makeID Crawlton		; 9E
	makeID Shellcracker		; 9F
	makeID ShellcrackerClaw	; A0
	makeID Slicer		; A1
	makeID SlicerPincers	; A2
	makeID Flasher		; A3
	makeID Asteron		; A4
	makeID Spiny		; A5
	makeID SpinyOnWall		; A6
	makeID Grabber		; A7
	makeID GrabberLegs		; A8
	makeID GrabberBox		; A9
	makeID GrabberString	; AA
	makeID Balkiry		; AC
	makeID CluckerBase		; AD
	makeID Clucker		; AE
	makeID MechaSonic		; AF
	makeID SegaScreen		; B0
	makeID SonicOnSegaScr	; B1
	makeID Tornado		; B2
	makeID Cloud		; B3
	makeID VPropeller		; B4
	makeID HPropeller		; B5
	makeID TiltingPlatform	; B6
	makeID VerticalLaser	; B7
	makeID WallTurret		; B8
	makeID Laser		; B9
	makeID WFZWheel		; BA
	makeID WFZShipFire		; BC
	makeID SmallMetalPform	; BD
	makeID LateralCannon	; BE
	makeID WFZStick		; BF
	makeID SpeedLauncher	; C0
	makeID BreakablePlating	; C1
	makeID Rivet		; C2
	makeID TornadoSmoke		; C3
	makeID TornadoSmoke2	; C4
	makeID WFZBoss		; C5
	makeID Eggman		; C6
	makeID Eggrobo		; C7
	makeID Crawl		; C8
	makeID TtlScrPalChanger	; C9
	makeID CutScene		; CA
	makeID EndingSeqClouds	; CB
	makeID EndingSeqTrigger	; CC
	makeID EndingSeqBird	; CD
	makeID EndingSeqSonic	; CE
	makeID EndingSeqTails	; CE
	makeID TornadoHelixes	; CF
	makeID CNZRectBlocks	; D2
	makeID BombPrize		; D3
	makeID CNZBigBlock		; D4
	makeID Elevator		; D5
	makeID PointPokey		; D6
	makeID Bumper		; D7
	makeID BonusBlock		; D8
	makeID Grab			; D9
	makeID RingPrize		; DC

; Object IDs
offset :=	Obj_Index_S1
ptrsize :=	4
idstart :=	1
; S1 Object IDs
	makeID LavaBall		; 14
	makeID GrassFire		; 19
	makeID Crabmeat		; 1F
	makeID Missile	; 23
	makeID GlassBlock		; 30
	makeID ChainStomp		; 31
	makeID GeyserMaker		; 4B
	makeID LavaGeyser		; 4D
	makeID Orbinaut		; 60
	makeID Gargoyle		; 62
	makeID LabyrinthConvey		; 63
	makeID Caterkiller		; 78
	makeID BuzzBomber
	makeID Jaws
	makeID Burrobot
	makeID Roller
	makeID Yadrin
	makeID Basaran

	include "musicids.asm"

; Sound IDs
offset :=	SoundIndex
ptrsize :=	3
idstart :=	MusID__End
idtype :=	"Snd"

SndID__First = idstart
	makeID Jump			; A0
	makeID Checkpoint		; A1
	makeID SpikeSwitch		; A2
	makeID Hurt			; A3
	makeID Skidding		; A4
	makeID BlockPush		; A5
	makeID HurtBySpikes		; A6
	makeID Sparkle		; A7
	makeID Beep			; A8
	makeID Bwoop		; A9
	makeID Splash		; AA
	makeID Swish		; AB
	makeID BossHit		; AC
	makeID InhalingBubble	; AD
	makeID ArrowFiring		; AE
	makeID LavaBall		; AE
	makeID Shield		; AF
	makeID LaserBeam		; B0
	makeID Zap			; B1
	makeID Drown		; B2
	makeID FireBurn		; B3
	makeID Bumper		; B4
	makeID Ring			; B5
	makeID RingRight		; B5
	makeID SpikesMove		; B6
	makeID Rumbling		; B7
	makeID Smash		; B9
	makeID DoorSlam		; BB
	makeID SpindashRelease	; BC
	makeID Hammer		; BD
	makeID Roll			; BE
	makeID ContinueJingle	; BF
	makeID CasinoBonus		; C0
	makeID Explosion		; C1
	makeID WaterWarning		; C2
	makeID EnterGiantRing	; C3
	makeID BossExplosion	; C4
	makeID TallyEnd		; C5
	makeID RingSpill		; C6
	makeID ChainRise		; C7
	makeID Flamethrower		; C8
	makeID Bonus		; C9
	makeID SpecStageEntry	; CA
	makeID SlowSmash		; CB
	makeID Spring		; CC
	makeID Blip			; CD
	makeID RingLeft		; CE
	makeID Signpost		; CF
	makeID CNZBossZap		; D0
	makeID Signpost2P		; D3
	makeID OOZLidPop		; D4
	makeID SlidingSpike		; D5
	makeID CNZElevator		; D6
	makeID PlatformKnock	; D7
	makeID BonusBumper		; D8
	makeID LargeBumper		; D9
	makeID Gloop		; DA
	makeID PreArrowFiring	; DB
	makeID Fire			; DC
	makeID ArrowStick		; DD
	makeID Helicopter		; DE
	makeID SuperTransform	; DF
	makeID SpindashRev		; E0
	makeID Rumbling2		; E1
	makeID CNZLaunch		; E2
	makeID Flipper		; E3
	makeID HTZLiftClick		; E4
	makeID Leaves		; E5
	makeID MegaMackDrop		; E6
	makeID DrawbridgeMove	; E7
	makeID QuickDoorSlam	; E8
	makeID DrawbridgeDown	; E9
	makeID LaserBurst		; EA
	makeID Scatter		; EB
	makeID LaserFloor		; EB
	makeID Teleport		; EC
	makeID Error		; ED
	makeID MechaSonicBuzz	; EE
	makeID LargeLaser		; EF
	makeID OilSlide		; F0
	makeID Grab
	makeID Land
	makeID Slide
	makeID Ping
	makeID PushBlock
	makeID BasaranFlap
	makeID _End			; F1
    if MOMPASS == 2
	if SndID__End > MusID_StopSFX
		fatal "You have too many SndPtrs. SndID__End ($\{SndID__End}) can't exceed MusID_StopSFX ($\{MusID_StopSFX})."
	endif
    endif

; Special sound IDs

MusID_StopSFX =		$F8			; F8
MusID_FadeOut =		$F9			; F9
SndID_SegaSound =	$FA			; FA
MusID_SpeedUp =		$FB			; FB
MusID_SlowDown =	$FC			; FC
MusID_Stop =		$FD			; FD
MusID_Pause =		$FE			; FE
MusID_Unpause =		$FF			; FF

; Music List IDs
offset :=	MusicList
ptrsize :=	2
idstart :=	0
idtype :=	"MusLst"

	makeID Title
	makeID Menu
	makeID EHZ
	makeID CPZ
	makeID ARZ
	makeID CNZ
	makeID HTZ
	makeID MCZ
	makeID OOZ
	makeID MTZ
	makeID WFZ
	makeID MZ
	makeID SYZ
	makeID LZ
	makeID HPZ
	makeID EHZ_2P
	makeID CNZ_2P
	makeID MCZ_2P
	makeID Invincibility
	makeID ActClear
	makeID GameOver
	makeID ExtraLife
	makeID Credits

; Animation IDs
offset :=	KnucklesAniData
ptrsize :=	2
idstart :=	0

AniIDKnuxAni_Walk			= id(KnuxAni_Walk_ptr)		; 0
AniIDKnuxAni_Run			= id(KnuxAni_Run_ptr)		; 1
AniIDKnuxAni_Roll			= id(KnuxAni_Roll_ptr)		; 2
AniIDKnuxAni_Roll2			= id(KnuxAni_Roll2_ptr)		; 3
AniIDKnuxAni_Push			= id(KnuxAni_Push_ptr)		; 4
AniIDKnuxAni_Wait			= id(KnuxAni_Wait_ptr)		; 5
AniIDKnuxAni_Balance			= id(KnuxAni_Balance_ptr)	; 6
AniIDKnuxAni_LookUp			= id(KnuxAni_LookUp_ptr)	; 7
AniIDKnuxAni_Duck			= id(KnuxAni_Duck_ptr)		; 8
AniIDKnuxAni_Spindash			= id(KnuxAni_Spindash_ptr)	; 9
AniIDKnuxAni_Unused			= id(KnuxAni_Unused_ptr)	; A
AniIDKnuxAni_Pull			= id(KnuxAni_Pull_ptr)		; B
AniIDKnuxAni_Balance2			= id(KnuxAni_Balance2_ptr)	; C
AniIDKnuxAni_Stop			= id(KnuxAni_Stop_ptr)		; D
AniIDKnuxAni_Float			= id(KnuxAni_Float_ptr)		; E
AniIDKnuxAni_Float2			= id(KnuxAni_Float2_ptr)	; F
AniIDKnuxAni_Spring			= id(KnuxAni_Spring_ptr)	; 10
AniIDKnuxAni_Hang			= id(KnuxAni_Hang_ptr)		; 11
AniIDKnuxAni_Unused_0			= id(KnuxAni_Unused_0_ptr)	; 12
AniIDKnuxAni_S3EndingPose			= id(KnuxAni_S3EndingPose_ptr)	; 13
AniIDKnuxAni_WFZHang			= id(KnuxAni_WFZHang_ptr)	; 14
AniIDKnuxAni_Bubble			= id(KnuxAni_Bubble_ptr)	; 15
AniIDKnuxAni_DeathBW			= id(KnuxAni_DeathBW_ptr)	; 16
AniIDKnuxAni_Drown			= id(KnuxAni_Drown_ptr)		; 17
AniIDKnuxAni_Death			= id(KnuxAni_Death_ptr)		; 18
AniIDKnuxAni_OilSlide			= id(KnuxAni_OilSlide_ptr)	; 19
AniIDKnuxAni_Hurt			= id(KnuxAni_Hurt_ptr)		; 1A
AniIDKnuxAni_OilSlide_0			= id(KnuxAni_OilSlide_0_ptr)	; 1B
AniIDKnuxAni_Blank			= id(KnuxAni_Blank_ptr)		; 1C
AniIDKnuxAni_Unused_1			= id(KnuxAni_Unused_1_ptr)	; 1D
AniIDKnuxAni_Unused_2			= id(KnuxAni_Unused_2_ptr)	; 1E
AniIDKnuxAni_Transform			= id(KnuxAni_Transform_ptr)	; 1F
AniIDKnuxAni_Gliding			= id(KnuxAni_Gliding_ptr)	; 20
AniIDKnuxAni_FallFromGlide			= id(KnuxAni_FallFromGlide_ptr)	; 21
AniIDKnuxAni_GetUp			= id(KnuxAni_GetUp_ptr)		; 22
AniIDKnuxAni_HardFall			= id(KnuxAni_HardFall_ptr)	; 23
AniIDKnuxAni_Badass			= id(KnuxAni_Badass_ptr)	; 24


offset :=	TailsAniData
ptrsize :=	2
idstart :=	0

AniIDTailsAni_Walk			= id(TailsAni_Walk_ptr)			;  0 ;   0
AniIDTailsAni_Run			= id(TailsAni_Run_ptr)			;  1 ;   1
AniIDTailsAni_Roll			= id(TailsAni_Roll_ptr)			;  2 ;   2
AniIDTailsAni_Roll2			= id(TailsAni_Roll2_ptr)		;  3 ;   3
AniIDTailsAni_Push			= id(TailsAni_Push_ptr)			;  4 ;   4
AniIDTailsAni_Wait			= id(TailsAni_Wait_ptr)			;  5 ;   5
AniIDTailsAni_Balance		= id(TailsAni_Balance_ptr)		;  6 ;   6
AniIDTailsAni_LookUp		= id(TailsAni_LookUp_ptr)		;  7 ;   7
AniIDTailsAni_Duck			= id(TailsAni_Duck_ptr)			;  8 ;   8
AniIDTailsAni_Spindash		= id(TailsAni_Spindash_ptr)		;  9 ;   9
AniIDTailsAni_Dummy1		= id(TailsAni_Dummy1_ptr)		; 10 ;  $A
AniIDTailsAni_Dummy2		= id(TailsAni_Dummy2_ptr)		; 11 ;  $B
AniIDTailsAni_Dummy3		= id(TailsAni_Dummy3_ptr)		; 12 ;  $C
AniIDTailsAni_Stop			= id(TailsAni_Stop_ptr)			; 13 ;  $D
AniIDTailsAni_Float			= id(TailsAni_Float_ptr)		; 14 ;  $E
AniIDTailsAni_Float2		= id(TailsAni_Float2_ptr)		; 15 ;  $F
AniIDTailsAni_Spring		= id(TailsAni_Spring_ptr)		; 16 ; $10
AniIDTailsAni_Hang			= id(TailsAni_Hang_ptr)			; 17 ; $11
AniIDTailsAni_Blink			= id(TailsAni_Blink_ptr)		; 18 ; $12
AniIDTailsAni_Blink2		= id(TailsAni_Blink2_ptr)		; 19 ; $13
AniIDTailsAni_Hang2			= id(TailsAni_Hang2_ptr)		; 20 ; $14
AniIDTailsAni_Bubble		= id(TailsAni_Bubble_ptr)		; 21 ; $15
AniIDTailsAni_DeathBW		= id(TailsAni_DeathBW_ptr)		; 22 ; $16
AniIDTailsAni_Drown			= id(TailsAni_Drown_ptr)		; 23 ; $17
AniIDTailsAni_Death			= id(TailsAni_Death_ptr)		; 24 ; $18
AniIDTailsAni_Hurt			= id(TailsAni_Hurt_ptr)			; 25 ; $19
AniIDTailsAni_Hurt2			= id(TailsAni_Hurt2_ptr)		; 26 ; $1A
AniIDTailsAni_Slide			= id(TailsAni_Slide_ptr)		; 27 ; $1B
AniIDTailsAni_Blank			= id(TailsAni_Blank_ptr)		; 28 ; $1C
AniIDTailsAni_Dummy4		= id(TailsAni_Dummy4_ptr)		; 29 ; $1D
AniIDTailsAni_Dummy5		= id(TailsAni_Dummy5_ptr)		; 30 ; $1E
AniIDTailsAni_HaulAss		= id(TailsAni_HaulAss_ptr)		; 31 ; $1F
AniIDTailsAni_Fly			= id(TailsAni_Fly_ptr)			; 32 ; $20


; Other sizes
palette_line_size =	$10	; 16 word entries

	enum rank_Locked,rank_Unplayed,rank_E,rank_D,rank_C,rank_B,rank_A

LevelData	STRUCT
Rank:	ds.b	1
PlayCount:	ds.b	1
Score:	ds.l	1
Time:
	ds.b	1
Time_Min:	ds.b	1
Time_Sec:	ds.b	1
Time_Fra:	ds.b	1
Rings:	ds.w	1
	ENDSTRUCT

	enum difficulty_Normal,difficulty_Hard,difficulty_Easy

SaveData	STRUCT
Player:	ds.b	1
Difficulty:	ds.b	1
P1Wins:	ds.b	1
P2Wins:	ds.b	1
Levels:
Levels_EHZ:	LevelData
Levels_CPZ:	LevelData
Levels_ARZ:	LevelData
Levels_CNZ:	LevelData
Levels_HTZ:	LevelData
Levels_MCZ:	LevelData
Levels_OOZ:	LevelData
Levels_MTZ:	LevelData
Levels_WFZ:	LevelData
Levels_MZ:	LevelData
Levels_SYZ:	LevelData
Levels_LZ:	LevelData
Levels_HPZ:	LevelData
MusicChoices:
Music_Title:	ds.b	1
Music_Menu:	ds.b	1
Music_Levels:
Music_EHZ:	ds.b	1
Music_CPZ:	ds.b	1
Music_ARZ:	ds.b	1
Music_CNZ:	ds.b	1
Music_HTZ:	ds.b	1
Music_MCZ:	ds.b	1
Music_OOZ:	ds.b	1
Music_MTZ:	ds.b	1
Music_WFZ:	ds.b	1
Music_MZ:	ds.b	1
Music_SYZ:	ds.b	1
Music_LZ:	ds.b	1
Music_HPZ:	ds.b	1
Music_EHZ_2P:	ds.b	1
Music_CNZ_2P:	ds.b	1
Music_MCZ_2P:	ds.b	1
Music_Invincibility:	ds.b	1
Music_ActClear:	ds.b	1
Music_GameOver:	ds.b	1
Music_ExtraLife:	ds.b	1
Music_Credits:	ds.b	1
	ds.b	1
	ENDSTRUCT

	if SaveData.len & 3
	fatal "Whoops, SaveData struct needs \{4 - (SaveData.len & 3)} bytes of padding to work."
	endif

; ---------------------------------------------------------------------------
; I run the main 68k RAM addresses through this function
; to let them work in both 16-bit and 32-bit addressing modes.
ramaddr function x,-(-x)&$FFFFFFFF

; ---------------------------------------------------------------------------
; RAM variables - General
	phase	ramaddr($FFFF0000)	; Pretend we're in the RAM
RAM_Start:

Chunk_Table:			ds.b	$8000	; was "Metablock_Table"
Chunk_Table_End:

Block_Table:			ds.w	$C00
Block_Table_End:

TempArray_LayerDef:		ds.b	$200	; used by some layer deformation routines
Decomp_Buffer:			ds.b	$200
Sprite_Table_Input:		ds.b	$400	; in custom format before being converted and stored in Sprite_Table/Sprite_Table_2
Sprite_Table_Input_End:

Object_RAM:			; The various objects in the game are loaded in this area.
				; Each game mode uses different objects, so some slots are reused.
				; The section below declares labels for the objects used in main gameplay.
				; Objects for other game modes are declared further down.
Reserved_Object_RAM:
MainCharacter:			; first object (usually Sonic except in a Tails Alone game)
				ds.b	object_size
Sidekick:			; second object (Tails in a Sonic and Tails game)
				ds.b	object_size
TitleCard:
TitleCard_ZoneName:		; level title card: zone name
GameOver_GameText:		; "GAME" from GAME OVER
TimeOver_TimeText:		; "TIME" from TIME OVER
				ds.b	object_size
TitleCard_Zone:			; level title card: "ZONE"
GameOver_OverText:		; "OVER" from GAME OVER
TimeOver_OverText:		; "OVER" from TIME OVER
				ds.b	object_size
TitleCard_Background:		; level title card: background
				ds.b	object_size
TitleCard_Bottom:		; level title card: yellow part at the bottom
				ds.b	object_size
TitleCard_Left:			; level title card: red part on the left
				ds.b	object_size


CPZPylon:			; Pylon in the foreground in CPZ
				ds.b	object_size
WaterSurface1:			; First water surface
Oil:				; Oil at the bottom of OOZ
				ds.b	object_size
WaterSurface2:			; Second water surface
				ds.b	object_size
Reserved_Object_RAM_End:

Dynamic_Object_RAM:		; Dynamic object RAM
				ds.b	$28*object_size
Dynamic_Object_RAM_2P_End:	; SingleObjLoad stops searching here in 2P mode
				ds.b	$71*object_size
Dynamic_Object_RAM_End:

LevelOnly_Object_RAM:
Sonic_BreathingBubbles:		; Sonic's breathing bubbles
				ds.b	object_size
Tails_BreathingBubbles:		; Tails' breathing bubbles
				ds.b	object_size
Sonic_Dust:			; Sonic's spin dash dust
				ds.b	object_size
Tails_Dust:			; Tails' spin dash dust
				ds.b	object_size
Sonic_Shield:
				ds.b	object_size
Tails_Shield:
				ds.b	object_size
Sonic_InvincibilityStars:
				ds.b	object_size
				ds.b	object_size
				ds.b	object_size
				ds.b	object_size
Tails_InvincibilityStars:
				ds.b	object_size
				ds.b	object_size
				ds.b	object_size
				ds.b	object_size
LevelOnly_Object_RAM_End:
Knuckles_EmeraldManager	EmeraldManager
Object_RAM_End:

Level_Layout:			ds.b	$1000
Level_Layout_End:

Primary_Collision:		ds.b	$300
Secondary_Collision:		ds.b	$300
VDP_Command_Buffer:		ds.w	7*$12	; stores 18 ($12) VDP commands to issue the next time ProcessDMAQueue is called
VDP_Command_Buffer_Slot:	ds.w	1	; stores the address of the next open slot for a queued VDP command

Sprite_Table_2:			ds.b	$280	; Sprite attribute table buffer for the bottom split screen in 2-player mode
Horiz_Scroll_Buf:		ds.b	$380
Horiz_Scroll_Buf_End:
Next_H_Int_Line:		ds.w	1
Sonic_Pos_Record_Buf:		ds.b	$100
Sonic_Pos_Record_Buf_End
Sonic_Stat_Record_Buf:		ds.b	$100
Tails_Pos_Record_Buf:		ds.b	$100
S1_BG_Scrolling:
CNZ_saucer_data:		ds.b	$40	; the number of saucer bumpers in a group which have been destroyed. Used to decide when to give 500 points instead of 10
Ring_Status_Table:			ds.w	768
Ring_Status_Table_End

Camera_RAM:
Camera_X_pos:			ds.l	1
Camera_Y_pos:			ds.l	1
Camera_BG_X_pos:		ds.l	1	; only used sometimes as the layer deformation makes it sort of redundant
Camera_BG_Y_pos:		ds.l	1
Camera_BG2_X_pos:		ds.l	1	; used in CPZ
Camera_BG2_Y_pos:		ds.l	1	; used in CPZ
Camera_BG3_X_pos:		ds.l	1	; unused (only initialised at beginning of level)?
Camera_BG3_Y_pos:		ds.l	1	; unused (only initialised at beginning of level)?
Camera_X_pos_P2:		ds.l	1
Camera_Y_pos_P2:		ds.l	1
Camera_BG_X_pos_P2:		ds.l	1	; only used sometimes as the layer deformation makes it sort of redundant
Camera_BG_Y_pos_P2:		ds.l	1
Camera_BG2_X_pos_P2:	ds.w	1	; unused (only initialised at beginning of level)?
Camera_BG2_Y_pos_P2:	ds.l	1
Camera_BG3_X_pos_P2:	ds.w	1	; unused (only initialised at beginning of level)?
Camera_BG3_Y_pos_P2:	ds.l	1
Horiz_block_crossed_flag:	ds.b	1	; toggles between 0 and $10 when you cross a block boundary horizontally
Verti_block_crossed_flag:	ds.b	1	; toggles between 0 and $10 when you cross a block boundary vertically
Horiz_block_crossed_flag_BG:	ds.b	1	; toggles between 0 and $10 when background camera crosses a block boundary horizontally
Verti_block_crossed_flag_BG:	ds.b	1	; toggles between 0 and $10 when background camera crosses a block boundary vertically
Horiz_block_crossed_flag_BG2:	ds.b	1	; used in CPZ
Horiz_block_crossed_flag_BG3:	ds.b	1
Horiz_block_crossed_flag_P2:	ds.b	1	; toggles between 0 and $10 when you cross a block boundary horizontally
Verti_block_crossed_flag_P2:	ds.b	1	; toggles between 0 and $10 when you cross a block boundary vertically
Scroll_flags:			ds.w	1	; bitfield ; bit 0 = redraw top row, bit 1 = redraw bottom row, bit 2 = redraw left-most column, bit 3 = redraw right-most column
Scroll_flags_BG:		ds.w	1	; bitfield ; bit 0-3 as above, bit 4-7 unknown (used by some deformation routines)
Scroll_flags_BG2:		ds.w	1	; used in CPZ; bit 0-1 unknown
Scroll_flags_BG3:		ds.w	1	; used in CPZ; bit 0-1 unknown
Scroll_flags_P2:		ds.w	1	; bitfield ; bit 0 = redraw top row, bit 1 = redraw bottom row, bit 2 = redraw left-most column, bit 3 = redraw right-most column
Scroll_flags_BG_P2:		ds.w	1	; bitfield ; bit 0-3 as above, bit 4-7 unknown (used by some deformation routines)
Scroll_flags_BG2_P2:	ds.w	1	; used in CPZ; bit 0-1 unknown
Scroll_flags_BG3_P2:	ds.w	1	; used in CPZ; bit 0-1 unknown
Camera_RAM_copy:		ds.l	2	; copied over every V-int
Camera_BG_copy:			ds.l	2	; copied over every V-int
Camera_BG2_copy:		ds.l	2	; copied over every V-int
Camera_BG3_copy:		ds.l	2	; copied over every V-int
Camera_P2_copy:			ds.l	8	; copied over every V-int
Scroll_flags_copy:		ds.w	1	; copied over every V-int
Scroll_flags_BG_copy:		ds.w	1	; copied over every V-int
Scroll_flags_BG2_copy:		ds.w	1	; copied over every V-int
Scroll_flags_BG3_copy:		ds.w	1	; copied over every V-int
Scroll_flags_copy_P2:		ds.w	1	; copied over every V-int
Scroll_flags_BG_copy_P2:	ds.w	1	; copied over every V-int
Scroll_flags_BG2_copy_P2:	ds.w	1	; copied over every V-int
Scroll_flags_BG3_copy_P2:	ds.w	1	; copied over every V-int

Camera_X_pos_diff:		ds.w	1	; (new X pos - old X pos) * 256
Camera_Y_pos_diff:		ds.w	1	; (new Y pos - old Y pos) * 256
Camera_BG_X_pos_diff:	ds.w	1	; Effective camera change used in WFZ ending and HTZ screen shake
Camera_BG_Y_pos_diff:	ds.w	1	; Effective camera change used in WFZ ending and HTZ screen shake
Camera_X_pos_diff_P2:		ds.w	1	; (new X pos - old X pos) * 256
Camera_Y_pos_diff_P2:		ds.w	1	; (new Y pos - old Y pos) * 256
Screen_Shaking_Flag_HTZ:	ds.b	1	; activates screen shaking code in HTZ's layer deformation routine
Screen_Shaking_Flag:		ds.b	1	; activates screen shaking code (if existent) in layer deformation routine
Scroll_lock:			ds.b	1	; set to 1 to stop all scrolling for P1
Scroll_lock_P2:			ds.b	1	; set to 1 to stop all scrolling for P2
Camera_Max_Y_pos:		ds.w	1
Camera_Min_X_pos:		ds.w	1
Camera_Max_X_pos:		ds.w	1
Camera_Min_Y_pos:		ds.w	1
Camera_Max_Y_pos_now:		ds.w	1	; was "Camera_max_scroll_spd"...
Horiz_scroll_delay_val:		ds.w	1	; if its value is a, where a != 0, X scrolling will be based on the player's X position a-1 frames ago
Sonic_Pos_Record_Index:		ds.w	1	; into Sonic_Pos_Record_Buf and Sonic_Stat_Record_Buf
Horiz_scroll_delay_val_P2:	ds.w	1
Tails_Pos_Record_Index:		ds.w	1	; into Tails_Pos_Record_Buf
Camera_Y_pos_bias:		ds.w	1	; added to y position for lookup/lookdown, $60 is center
Camera_Y_pos_bias_P2:		ds.w	1	; for Tails
Deform_lock:			ds.b	1	; set to 1 to stop all deformation
Camera_Max_Y_Pos_Changing:	ds.b	1
Dynamic_Resize_Routine:		ds.b	1
Use_Alt_Pal_Mus:	ds.b	1
unk_EEE2:			ds.w	1	; Used to control background scrolling in X in WFZ ending and HTZ screen shake
unk_EEE4:			ds.w	1	; Used to control background scrolling in Y in WFZ ending and HTZ screen shake
unk_EEE6:			ds.w	1	; Used as some sort of timer in HTZ2 screen shake
unk_EEE8:			ds.b	1	; Flag of some sort to control vertical direction for screen shake in HTZ2
	ds.b	1 ; filler
Vscroll_Factor_P2_HInt:	ds.l	1
Camera_X_pos_copy:		ds.l	1
Camera_Y_pos_copy:		ds.l	1
Tails_Min_X_pos:		ds.w	1
Tails_Max_X_pos:		ds.w	1
Tails_Min_Y_pos:		ds.w	1 ; seems not actually implemented (only written to)
Tails_Max_Y_pos:		ds.w	1
Camera_RAM_End:

Block_cache:			ds.b	$80
Ring_consumption_table:		ds.b	$80	; contains RAM addresses of rings currently being consumed
Ring_consumption_table_End

Underwater_palette_2:		ds.w palette_line_size	; not sure what it's used for but it's only used when there's water
Underwater_palette_2_line2:	ds.w palette_line_size
Underwater_palette_2_line3:	ds.w palette_line_size
Underwater_palette_2_line4:	ds.w palette_line_size

Underwater_palette:		ds.w palette_line_size	; main palette for underwater parts of the screen
Underwater_palette_line2:	ds.w palette_line_size
Underwater_palette_line3:	ds.w palette_line_size
Underwater_palette_line4:	ds.w palette_line_size

Palette_Buffer:			ds.b	$16
LZ_Water_Ripple:		ds.w	1
Palette_Buffer_End

Game_Mode:			ds.w	1	; 1 byte ; see GameModesArray (master level trigger, Mstr_Lvl_Trigger)
Ctrl_1_Logical:					; 2 bytes
Ctrl_1_Held_Logical:		ds.b	1	; 1 byte
Ctrl_1_Press_Logical:		ds.b	1	; 1 byte
Ctrl_1:						; 2 bytes
Ctrl_1_Held:			ds.b	1	; 1 byte ; (pressed and held were switched around before)
Ctrl_1_Press:			ds.b	1	; 1 byte
Ctrl_2:						; 2 bytes
Ctrl_2_Held:			ds.b	1	; 1 byte
Ctrl_2_Press:			ds.b	1	; 1 byte
VDP_Reg1_val:			ds.w	1	; normal value of VDP register #1 when display is disabled
Demo_Time_left:			ds.w	1	; 2 bytes

Vscroll_Factor:
Vscroll_Factor_FG:			ds.w	1
Vscroll_Factor_BG:			ds.w	1
unk_F61A:			ds.l	1	; Only ever cleared, never used
Vscroll_Factor_P2:
Vscroll_Factor_P2_FG:		ds.w	1
Vscroll_Factor_P2_BG:		ds.w	1
Hint_counter_reserve:		ds.w	1	; Must contain a VDP command word, preferably a write to register $0A. Executed every V-INT.
Palette_fade_range:				; Range affected by the palette fading routines
Palette_fade_start:		ds.b	1	; Offset from the start of the palette to tell what range of the palette will be affected in the palette fading routines
Palette_fade_length:		ds.b	1	; Number of entries to change in the palette fading routines

MiscLevelVariables:
				ds.b	1	; unused
Vint_routine:			ds.b	1	; was "Delay_Time" ; routine counter for V-int
Sprite_count:			ds.b	1	; the number of sprites drawn in the current frame
ShowScoreStar:	ds.b	1
ShowTimeStar:	ds.b	1
ShowRingsStar:	ds.b	1
PalCycle_Frame:			ds.w	1	; ColorID loaded in PalCycle
PalCycle_Timer:			ds.w	1	; number of frames until next PalCycle call
RNG_seed:			ds.l	1	; used for random number generation
Game_paused:			ds.w	1	
Knuckles_GlideSomething:	ds.b	1
Tikal_GlideSomething:	ds.b	1
DMA_data_thunk:			ds.w	1	; Used as a RAM holder for the final DMA command word. Data will NOT be preserved across V-INTs, so consider this space reserved.
Hint_flag:			ds.w	1	; unless this is 1, H-int won't run

Water_Level_1:			ds.w	1
Water_Level_2:			ds.w	1
Water_Level_3:			ds.w	1
Water_on:			ds.b	1	; is set based on Water_flag
Water_routine:			ds.b	1
Water_fullscreen_flag:		ds.b	1	; was "Water_move"
Do_Updates_in_H_int:	ds.b	1

PalCycle_Frame_CNZ:		ds.w	1
PalCycle_Frame2:		ds.w	1
PalCycle_Frame3:		ds.w	1
PalCycle_Frame2_CNZ:	ds.w	1
Palette_frame:			ds.w	1
Palette_timer:			ds.b	1	; was "Palette_frame_count"
Super_Sonic_palette:		ds.b	1

DEZ_Eggman:						; Word
DEZ_Shake_Timer:				; Word
WFZ_LevEvent_Subrout:			; Word
SegaScr_PalDone_Flag:			; Byte (cleared once as a word)
Credits_Trigger:		ds.b	1	; cleared as a word a couple times
Ending_PalCycle_flag:	ds.b	1

SegaScr_VInt_Subrout:
Ending_VInt_Subrout:
WFZ_Event_Counter:		ds.w	1
PalCycle_Timer2:		ds.w	1
PalCycle_Timer3:		ds.w	1

Ctrl_2_Logical:					; 2 bytes
Ctrl_2_Held_Logical:		ds.b	1	; 1 byte
Ctrl_2_Press_Logical:		ds.b	1	; 1 byte
Sonic_Look_delay_counter:	ds.w	1	; 2 bytes
Tails_Look_delay_counter:	ds.w	1	; 2 bytes
Super_Sonic_frame_count:	ds.w	1
Camera_ARZ_BG_X_pos:		ds.l	1
CNZ_Point_Slot_Uses:		ds.b	1
CNZ_Ring_Slot_Uses:			ds.b	1
MiscLevelVariables_End

Plc_Buffer:			ds.b	$60	; Pattern load queue (each entry is 6 bytes)
Plc_Buffer_Only_End:
				; these seem to store nemesis decompression state so PLC processing can be spread out across frames
Plc_Buffer_Reg0:		ds.l	1	
Plc_Buffer_Reg4:		ds.l	1	
Plc_Buffer_Reg8:		ds.l	1	
Plc_Buffer_RegC:		ds.l	1	
Plc_Buffer_Reg10:		ds.l	1	
Plc_Buffer_Reg14:		ds.l	1	
Plc_Buffer_Reg18:		ds.w	1	; amount of current entry remaining to decompress
Plc_Buffer_Reg1A:		ds.w	1	
Plc_Buffer_End:


Misc_Variables:
Obj31_S1_Y_pos:			ds.w	1

Rings_manager_routine:		ds.b	1
Level_started_flag:		ds.b	1
Ring_start_addr:		ds.l	1
Ring_end_addr:			ds.l	1
Ring_status_addr:		ds.w	1
Ring_start_addr_P2:		ds.l	1
Ring_end_addr_P2:		ds.l	1
Ring_status_addr_P2:	ds.w	1
CNZ_Bumper_routine:		ds.b	1
CNZ_Bumper_UnkFlag:		ds.b	1	; Set only, never used again
CNZ_Visible_bumpers_start:			ds.l	1
CNZ_Visible_bumpers_end:			ds.l	1
CNZ_Visible_bumpers_start_P2:			ds.l	1
CNZ_Visible_bumpers_end_P2:			ds.l	1

Dirty_flag:			ds.b	1	; if whole screen needs to redraw
CPZ_UnkScroll_Timer:	ds.b	1	; Used only in unused CPZ scrolling function
WFZ_SCZ_Fire_Toggle:	ds.b	1
Water_flag:			ds.b	1	; if the level has water or oil
Demo_button_index_2P:		ds.w	1	; index into button press demo data, for player 2
Demo_press_counter_2P:		ds.w	1	; frames remaining until next button press, for player 2
Tornado_Velocity_X:		ds.w	1	; speed of tails' plane in scz ($FFFFF736)
Tornado_Velocity_Y:		ds.w	1
ScreenShift:			ds.b	1
Boss_CollisionRoutine:		ds.b	1
Boss_AnimationArray:		ds.b	$10	; up to $10 bytes; 2 bytes per entry
Ending_Routine:
Boss_X_pos:			ds.w	1
				ds.w	1	; Boss_MoveObject reads a long, but all other places in the game use only the high word
Boss_Y_pos:			ds.w	1
				ds.w	1	; same here
Boss_X_vel:			ds.w	1
Boss_Y_vel:			ds.w	1
Boss_Countdown:		ds.w	1

Sonic_top_speed:		ds.w	1
Sonic_acceleration:		ds.w	1
Sonic_deceleration:		ds.w	1
Sonic_LastLoadedDPLC:		ds.b	1	; mapping frame number when Sonic last had his tiles requested to be transferred from ROM to VRAM. can be set to a dummy value like -1 to force a refresh DMA. was: Sonic_mapping_frame
Primary_Angle:		ds.b	1
Secondary_Angle:	ds.b	1
Obj_placement_routine:		ds.b	1
Camera_X_pos_last:		ds.w	1	; Camera_X_pos_coarse from the previous frame

Obj_load_addr_right:		ds.l	1	; contains the address of the next object to load when moving right
Obj_load_addr_left:		ds.l	1	; contains the address of the last object loaded when moving left
Obj_load_addr_2:		ds.l	1
Obj_load_addr_3:		ds.l	1
unk_F780:			ds.b	6	; seems to be an array of horizontal chunk positions, used for object position range checks
unk_F786:			ds.b	3
unk_F789:			ds.b	3
Camera_X_pos_last_P2:		ds.w	1
Obj_respawn_index_P2:		ds.b	2	; respawn table indices of the next objects when moving left or right for the second player

Demo_button_index:		ds.w	1	; index into button press demo data, for player 1
PalChangeSpeed:			ds.w	1
Collision_addr:			ds.l	1
ColCurveMap:			ds.l	1
ColArray:				ds.l	1
ColArray2:				ds.l	1
Boss_defeated_flag:		ds.b	1
Current_Boss_ID:		ds.b	1
MTZ_Platform_Cog_X:			ds.w	1	; X position of moving MTZ platform for cog animation.
MTZCylinder_Angle_Sonic:	ds.b	1
MTZCylinder_Angle_Tails:	ds.b	1
BigRingGraphics:	ds.w	1	; S1 holdover
WindTunnel_flag:		ds.b	1
WindTunnel_holding_flag:			ds.b	1
Control_Locked:			ds.b	1
SpecialStage_flag_2P:			ds.b	1
Control_Locked_P2:		ds.b	1
Conveyors_Reversed:		ds.b	1
Chain_Bonus_counter:		ds.w	1	; counts up when you destroy things that give points, resets when you touch the ground
Bonus_Countdown_1:		ds.w	1	; level results time bonus or special stage sonic ring bonus
Bonus_Countdown_2:		ds.w	1	; level results ring bonus or special stage tails ring bonus
Update_Bonus_score:		ds.b	1
Obj56_S1_Variable:		ds.b	1
Camera_X_pos_coarse:		ds.w	1	; (Camera_X_pos - 128) / 256
Camera_X_pos_coarse_P2:		ds.w	1
Tails_LastLoadedDPLC:		ds.b	1	; mapping frame number when Tails last had his tiles requested to be transferred from ROM to VRAM. can be set to a dummy value like -1 to force a refresh DMA.
TailsTails_LastLoadedDPLC:	ds.b	1	; mapping frame number when Tails' tails last had their tiles requested to be transferred from ROM to VRAM. can be set to a dummy value like -1 to force a refresh DMA.
ButtonVine_Trigger:		ds.b	$10	; 16 bytes flag array, #subtype byte set when button/vine of respective subtype activated
Anim_Counters:			ds.b	$10	; $FFFFF7F0-$FFFFF7FF
Obj63_S1_Variables:		ds.b	6
Obj6B_S1_Variable:		ds.b	1
Demo_press_counter:		ds.b	1	; frames remaining until next button press, for player 1
Misc_Variables_End:

Sprite_Table:			ds.b	$280	; Sprite attribute table buffer

Normal_palette:			ds.w	palette_line_size
Normal_palette_line2:		ds.w	palette_line_size
Normal_palette_line3:		ds.w	palette_line_size
Normal_palette_line4:		ds.w	palette_line_size
Second_palette:
Target_palette:			ds.w	palette_line_size
Second_palette_line2:
Target_palette_line2:		ds.w	palette_line_size
Second_palette_line3:
Target_palette_line3:		ds.w	palette_line_size
Second_palette_line4:
Target_palette_line4:		ds.w	palette_line_size

Object_Respawn_Table:
Obj_respawn_index:		ds.b	2		; respawn table indices of the next objects when moving left or right for the first player
Obj_respawn_data:		ds.b	$100	; Maximum possible number of respawn entries that S2 can handle; for stock S2, $80 is enough
Obj_respawn_data_End:
				ds.b	$FE	; Stack; at least $A0 bytes are needed.
System_Stack:

Level_Inactive_flag:		ds.w	1	; (2 bytes)
Timer_frames:			ds.w	1	; (2 bytes)
Debug_object:			ds.b	1
Life_count:			ds.b	1
Debug_placement_mode:		ds.b	1
				ds.b	1	; the whole word is tested, but the debug mode code uses only the low byte
Debug_Accel_Timer:	ds.b	1
Debug_Speed:		ds.b	1
Random_Counter:
	if Netplay==1
					ds.l	1
	endif
Vint_runcount:			ds.l	1

Current_Zone:			ds.b	1	; 1 byte
SRAM_Enabled:	ds.b	1
Super_Sonic_flag:		ds.b	1
Extra_life_flags:		ds.b	1

; If set, the respective HUD element will be updated.
Update_HUD_lives:		ds.b	1
Update_HUD_rings:		ds.b	1
Update_HUD_timer:		ds.b	1
Update_HUD_score:		ds.b	1
Timer:						; 4 bytes
Timer_minute_word:				; 2 bytes
				ds.b	1	; filler
Timer_minute:			ds.b	1	; 1 byte
Timer_second:			ds.b	1	; 1 byte
Timer_centisecond:				; inaccurate name (the seconds increase when this reaches 60)
Timer_frame:			ds.b	1	; 1 byte

Score:				ds.l	1	; 4 bytes
Saved_Timer:			ds.l	1
Oscillating_Numbers:
Oscillation_Control:			ds.w	1
Oscillating_variables:
Oscillating_Data:				ds.w	$20
Oscillating_Numbers_End

Logspike_anim_counter:		ds.b	1
Logspike_anim_frame:		ds.b	1
Rings_anim_counter:		ds.b	1
Rings_anim_frame:		ds.b	1
Unknown_anim_counter:		ds.b	1	; I think this was $FFFFFEC4 in the alpha
Unknown_anim_frame:		ds.b	1
Ring_spill_anim_counter:	ds.b	1	; scattered rings
Ring_spill_anim_frame:		ds.b	1
Ring_spill_anim_accum:		ds.w	1
Oscillating_variables_End

; values for the second player (some of these only apply to 2-player games)
Tails_top_speed:		ds.w	1	; Tails_max_vel
Tails_acceleration:		ds.w	1
Tails_deceleration:		ds.w	1
Life_count_2P:			ds.b	1
Extra_life_flags_2P:		ds.b	1
Update_HUD_lives_2P:		ds.b	1
Update_HUD_rings_2P:		ds.b	1
Update_HUD_timer_2P:		ds.b	1
Update_HUD_score_2P:		ds.b	1	; mostly unused
Ring_count_2P:			ds.w	1
Timer_2P:					; 4 bytes
Timer_minute_word_2P:				; 2 bytes
				ds.b	1	; filler
Timer_minute_2P:		ds.b	1	; 1 byte
Timer_second_2P:		ds.b	1	; 1 byte
Timer_centisecond_2P:				; inaccurate name (the seconds increase when this reaches 60)
Timer_frame_2P:			ds.b	1	; 1 byte
Score_2P:			ds.l	1
Last_star_pole_hit_2P:		ds.b	1
Saved_Last_star_pole_hit_2P:	ds.b	1
Saved_x_pos_2P:			ds.w	1
Saved_y_pos_2P:			ds.w	1
Saved_Ring_count_2P:		ds.w	1
Saved_Timer_2P:			ds.l	1
Saved_art_tile_2P:		ds.w	1
Saved_layer_2P:			ds.w	1
Perfect_rings_left:		ds.w	1
Perfect_rings_flag:			ds.w	1

Level_select_current_pal:
CreditsLineIndex:
SlotMachineInUse:			ds.w	1
SlotMachineVariables:	; $12 values
SlotMachine_Routine:	ds.b	1
SlotMachine_Timer:		ds.b	1
				ds.b	1	; $FFFFFF50 ; seems unused except for 1 write
SlotMachine_Index:		ds.b	1
SlotMachine_Reward:		ds.w	1
SlotMachine_Slot1Pos:	ds.w	1
SlotMachine_Slot1Speed:	ds.b	1
SlotMachine_Slot1Rout:	ds.b	1
SlotMachine_Slot2Pos:	ds.w	1
SlotMachine_Slot2Speed:	ds.b	1
SlotMachine_Slot2Rout:	ds.b	1
SlotMachine_Slot3Pos:	ds.w	1
SlotMachine_Slot3Speed:	ds.b	1
SlotMachine_Slot3Rout:	ds.b	1

Player_mode:			ds.w	1	; 0 = Sonic and Tails, 1 = Sonic, 2 = Tails
Player_option:			ds.w	1	; 0 = Sonic and Tails, 1 = Sonic, 2 = Tails

Difficulty:		ds.w	1
Level_ResultsActive:
LevSel_HoldTimer:		ds.w	1
Level_select_zone:		ds.w	1
Level_select_scroll_target:	ds.w	1
Level_select_scroll_speed:	ds.w	1
Sound_test_sound:		ds.w	1
Title_screen_option:		ds.b	1
Current_Zone_2P:		ds.b	1
Current_Act_2P:			ds.b	1
Options_menu_box:		ds.b	1
Music_switch_selection:		ds.b	1
Music_switch_scroll:		ds.b	1
Next_Rank_Counter:		ds.w	1
				
Level_Music:			ds.w	1
Ring_count:			ds.w	1	; 2 bytes
Bonus_Countdown_3:		ds.w	1
Game_Over_2P:			ds.w	1
SS2p_RingBuffer:		ds.w	6
SaveData_RAM:	SaveData
SaveData_RAM_End:
Next_Extra_life_score:		ds.l	1
Next_Extra_life_score_2P:	ds.l	1
Camera_Min_Y_pos_Debug_Copy:	ds.w	1
Camera_Max_Y_pos_Debug_Copy:	ds.w	1
Level_select_flag:		ds.b	1
Slow_motion_flag:		ds.b	1
Night_mode_flag:		ds.w	1
Correct_cheat_entries:		ds.w	1
Correct_cheat_entries_2:	ds.w	1	; for 14 continues or 7 emeralds codes
Two_player_mode:		ds.w	1	; flag (0 for main game)
unk_FFDC:			ds.b	1	; Written to near loc_175EA, never read from
unk_FFDD:			ds.b	1	; Written to near loc_175EA, never read from
unk_FFDE:			ds.b	1	; Written to near loc_175EA, never read from
unk_FFDF:			ds.b	1	; Written to near loc_175EA, never read from

; Values in these variables are passed to the sound driver during V-INT.
; They use a playlist index, not a sound test index.
Music_to_play:			ds.b	1
SFX_to_play:			ds.b	1	; normal
SFX_to_play_2:			ds.b	1	; alternating stereo
unk_FFE3:			ds.b	1
Music_to_play_2:		ds.b	1	; alternate (higher priority?) slot
Split_palette_enabled:		ds.b	1	;new line	;FFFFFFE6
Split_palette_Y:		ds.w	1	;new line
Alt_Object_Palette:		ds.w	3	;new line
Split_palette_mode:		ds.b	1	;new line	0=normal palette fullscreen, 1=splitscreen, 2=alt palette fullscreen
Art_Swap_ID:			ds.b	1	;new line
Demo_mode_flag:			ds.w	1 ; 1 if a demo is playing (2 bytes)
Demo_number:			ds.w	1 ; which demo will play next (2 bytes)
Graphics_Flags:			ds.w	1 ; misc. bitfield
Debug_mode_flag:		ds.w	1 ; (2 bytes)
Cur_Obj_Index:			ds.l	1
Sandwich_mode:			ds.w	1
Checksum_fourcc:		ds.l	1 ; (4 bytes)
RAM_End

    if * > 0	; Don't declare more space than the RAM can contain!
	fatal "The RAM variable declarations are too large by $\{*} bytes."
    endif

	if * < 0
	message "There are $\{-*} bytes of RAM free."
	endif

; RAM variables - SEGA screen
	phase	Object_RAM	; Move back to the object RAM
SegaScr_Object_RAM:
				; Unused slot
				ds.b	object_size
SegaScreenObject:		; Sega screen
				ds.b	object_size
SonicOnSegaScreen:		; Sonic on Sega screen
				ds.b	object_size

				ds.b	($80-3)*object_size
SegaScr_Object_RAM_End:


; RAM variables - Title screen
	phase	Object_RAM	; Move back to the object RAM
TtlScr_Object_RAM:
				; Unused slot
				ds.b	object_size
IntroSonic:			; stars on the title screen
				ds.b	object_size
IntroTails:
				ds.b	object_size
IntroLargeStar:
TitleScreenPaletteChanger:
				ds.b	object_size
TitleScreenPaletteChanger3:
				ds.b	object_size
IntroEmblemTop:
				ds.b	object_size
IntroSmallStar1:
				ds.b	object_size
IntroSonicHand:
				ds.b	object_size
IntroTailsHand:
				ds.b	object_size
TitleScreenPaletteChanger2:
				ds.b	object_size

				ds.b	6*object_size

TitleScreenMenu:
				ds.b	object_size
IntroSmallStar2:
				ds.b	object_size

				ds.b	($70-2)*object_size
TtlScr_Object_RAM_End:


; RAM variables - 2P VS results screen
	phase	Object_RAM	; Move back to the object RAM
VSRslts_Object_RAM:
VSResults_HUD:			; Blinking text at the bottom of the screen
				ds.b	object_size

				; Free slots
				ds.b	$7F*object_size
VSRslts_Object_RAM_End:


; RAM variables - Menu screens
	phase	Object_RAM	; Move back to the object RAM
Menus_Object_RAM:		; No objects are loaded in the menu screens
				ds.b	$80*object_size
Menus_Object_RAM_End:


; RAM variables - Ending sequence
	phase	Object_RAM
EndSeq_Object_RAM:
				ds.b	object_size
				ds.b	object_size
Tails_Tails_Cutscene:		; Tails' tails on the cut scene
				ds.b	object_size
EndSeqPaletteChanger:
				ds.b	object_size
CutScene:
				ds.b	object_size
				ds.b	($80-5)*object_size
EndSeq_Object_RAM_End:

; SRAM
SRAM_Access_Flag =			$A130F1
SRAM_Start =				$200001

	phase	SRAM_Start+2
SRAM_Magic:	ds.l	2
SRAM_Version:	ds.w	2
SRAM_PlayTime:	ds.l	2
SRAM_Checksum:	ds.w	2
SRAM_SaveData:	ds.b	SaveData.len*2

	dephase		; Stop pretending

	!org	0	; Reset the program counter


; ---------------------------------------------------------------------------
; VDP addressses
VDP_data_port =			$C00000 ; (8=r/w, 16=r/w)
VDP_control_port =		$C00004 ; (8=r/w, 16=r/w)
PSG_input =			$C00011

; ---------------------------------------------------------------------------
; Z80 addresses
Z80_RAM =			$A00000 ; start of Z80 RAM
Z80_RAM_End =			$A02000 ; end of non-reserved Z80 RAM
Z80_Bus_Request =		$A11100
Z80_Reset =			$A11200

Security_Addr =			$A14000

; ---------------------------------------------------------------------------
; I/O Area 
HW_Version =				$A10001
HW_Port_1_Data =			$A10003
HW_Port_2_Data =			$A10005
HW_Expansion_Data =			$A10007
HW_Port_1_Control =			$A10009
HW_Port_2_Control =			$A1000B
HW_Expansion_Control =		$A1000D
HW_Port_1_TxData =			$A1000F
HW_Port_1_RxData =			$A10011
HW_Port_1_SCtrl =			$A10013
HW_Port_2_TxData =			$A10015
HW_Port_2_RxData =			$A10017
HW_Port_2_SCtrl =			$A10019
HW_Expansion_TxData =		$A1001B
HW_Expansion_RxData =		$A1001D
HW_Expansion_SCtrl =		$A1001F

; ---------------------------------------------------------------------------
; Art tile stuff
flip_x_bit          =      3
flip_y_bit          =      4
flip_x              =      (1<<11)
flip_y              =      (1<<12)
palette_bit_0       =      5
palette_bit_1       =      6
palette_line_0      =      (0<<13)
palette_line_1      =      (1<<13)
palette_line_2      =      (2<<13)
palette_line_3      =      (3<<13)
high_priority_bit   =      7
high_priority       =      (1<<15)
palette_mask        =      $6000
tile_mask           =      $07FF
nontile_mask        =      $F800
drawing_mask        =      $7FFF

; ---------------------------------------------------------------------------
; VRAM and tile art base addresses.
; VRAM Reserved regions.
VRAM_Plane_A_Name_Table                  = $C000	; Extends until $CFFF
VRAM_Plane_B_Name_Table                  = $E000	; Extends until $EFFF
VRAM_Plane_A_Name_Table_2P               = $A000	; Extends until $AFFF
VRAM_Plane_B_Name_Table_2P               = $8000	; Extends until $8FFF
VRAM_Plane_Table_Size                    = $1000	; 64 cells x 32 cells x 2 bytes per cell
VRAM_Sprite_Attribute_Table              = $F800	; Extends until $FA7F
VRAM_Sprite_Attribute_Table_Size         = $0280	; 640 bytes
VRAM_Horiz_Scroll_Table                  = $FC00	; Extends until $FF7F
VRAM_Horiz_Scroll_Table_Size             = $0380	; 224 lines * 2 bytes per entry * 2 PNTs

; VRAM Reserved regions, Sega screen.
VRAM_SegaScr_Plane_A_Name_Table          = $C000	; Extends until $DFFF
VRAM_SegaScr_Plane_B_Name_Table          = $A000	; Extends until $BFFF
VRAM_SegaScr_Plane_Table_Size            = $2000	; 128 cells x 32 cells x 2 bytes per cell
VRAM_SegaScr_Sprite_Attribute_Table      = $F800	; Extends until $FA7F
VRAM_SegaScr_Sprite_Attribute_Table_Size = $0280	; 640 bytes
VRAM_SegaScr_Horiz_Scroll_Table          = $FC00	; Extends until $FF7F
VRAM_SegaScr_Horiz_Scroll_Table_Size     = $0380	; 224 lines * 2 bytes per entry * 2 PNTs

; VRAM Reserved regions, Title screen.
VRAM_TtlScr_Plane_A_Name_Table           = $C000	; Extends until $CFFF
VRAM_TtlScr_Plane_B_Name_Table           = $E000	; Extends until $EFFF
VRAM_TtlScr_Plane_Table_Size             = $1000	; 64 cells x 32 cells x 2 bytes per cell
VRAM_TtlScr_Sprite_Attribute_Table       = $F800	; Extends until $FA7F
VRAM_TtlScr_Sprite_Attribute_Table_Size  = $0280	; 640 bytes
VRAM_TtlScr_Horiz_Scroll_Table           = $FC00	; Extends until $FF7F
VRAM_TtlScr_Horiz_Scroll_Table_Size      = $0380	; 224 lines * 2 bytes per entry * 2 PNTs

; VRAM Reserved regions, Ending sequence and credits.
VRAM_EndSeq_Plane_A_Name_Table           = $C000	; Extends until $DFFF
VRAM_EndSeq_Plane_B_Name_Table           = $4000	; Extends until $5FFF
VRAM_EndSeq_Plane_Table_Size             = $2000	; 64 cells x 64 cells x 2 bytes per cell
VRAM_EndSeq_Sprite_Attribute_Table       = $F800	; Extends until $FA7F
VRAM_EndSeq_Sprite_Attribute_Table_Size  = $0280	; 640 bytes
VRAM_EndSeq_Horiz_Scroll_Table           = $FC00	; Extends until $FF7F
VRAM_EndSeq_Horiz_Scroll_Table_Size      = $0380	; 224 lines * 2 bytes per entry * 2 PNTs

; From here on, art tiles are used; VRAM address is art tile * $20.
ArtTile_VRAM_Start                    = $0000

; Common to 1p and 2p menus.
ArtTile_ArtNem_FontStuff              = $0010
ArtTile_MapEng_MenuBack               = $0700

; Sega screen
ArtTile_ArtNem_Sega_Logo              = $0001
ArtTile_ArtNem_Trails                 = $0080
ArtTile_ArtUnc_Giant_Sonic            = $0088

; Title screen
ArtTile_ArtNem_Title                  = $0010
ArtTile_ArtNem_TitlePlane             = $00F2
ArtTile_ArtNem_MenuJunk               = $03F2
ArtTile_ArtNem_Player1VS2             = $0402
ArtTile_ArtNem_CreditText             = $0500
ArtTile_ArtNem_FontStuff_TtlScr       = $0680

; Credits screen
ArtTile_ArtNem_CreditText_CredScr     = $0200

; Menu background.
ArtTile_ArtNem_MenuBox                = $0070

; Level select icons.
ArtTile_ArtNem_LevelSelectPics        = $0090

; Congratulations! screen.
ArtTile_ArtNem_CongratulationsRainbow = $0070
ArtTile_ArtNem_CongratulationsText    = ArtTile_ArtNem_CongratulationsRainbow+$49

; 2p results screen.
ArtTile_ArtNem_1P2PWins               = $0070
ArtTile_ArtNem_SpecialPlayerVSPlayer  = $03DF
ArtTile_ArtNem_2p_Signpost            = $05E8
ArtTile_TwoPlayerResults              = $0600

; Ending.
ArtTile_EndingCharacter               = $0019
ArtTile_ArtNem_EndingFinalTornado     = $0156
ArtTile_ArtNem_EndingPics             = $0328
ArtTile_ArtNem_EndingMiniTornado      = $0493

; S1 Ending
ArtTile_ArtNem_S1EndFlicky            = $05A5
ArtTile_ArtNem_S1EndRabbit            = $0553
ArtTile_ArtNem_S1EndPenguin           = $0573
ArtTile_ArtNem_S1EndSeal              = $0585
ArtTile_ArtNem_S1EndPig               = $0593
ArtTile_ArtNem_S1EndChicken           = $0565
ArtTile_ArtNem_S1EndSquirrel          = $05B3

; ---------------------------------------------------------------------------
; Level art stuff.
ArtTile_ArtKos_LevelArt               = $0000
ArtTile_ArtKos_NumTiles_EHZ           = $0393
ArtTile_ArtKos_NumTiles_CPZ           = $0364
ArtTile_ArtKos_NumTiles_ARZ           = $03F6
ArtTile_ArtKos_NumTiles_CNZ           = $0331
ArtTile_ArtKos_NumTiles_HTZ_Main      = $01FC ; Until this tile, equal to EHZ tiles.
ArtTile_ArtKos_NumTiles_HTZ_Sup       = $0183 ; Overwrites several EHZ tiles.
ArtTile_ArtKos_NumTiles_HTZ           = ArtTile_ArtKos_NumTiles_HTZ_Main + ArtTile_ArtKos_NumTiles_HTZ_Sup - 1
ArtTile_ArtKos_NumTiles_MCZ           = $03A9
ArtTile_ArtKos_NumTiles_OOZ           = $02AA
ArtTile_ArtKos_NumTiles_MTZ           = $0319
ArtTile_ArtKos_NumTiles_SCZ           = $036E
ArtTile_ArtKos_NumTiles_WFZ_Main      = $0307 ; Until this tile, equal to SCZ tiles.
ArtTile_ArtKos_NumTiles_WFZ_Sup       = $0073 ; Overwrites several SCZ tiles.
ArtTile_ArtKos_NumTiles_WFZ           = ArtTile_ArtKos_NumTiles_WFZ_Main + ArtTile_ArtKos_NumTiles_WFZ_Sup - 1
ArtTile_ArtKos_NumTiles_DEZ           = $0326 ; Skips several CPZ tiles.

; ---------------------------------------------------------------------------
; Shared badniks and objects.

; Objects that use the same art tiles on all levels in which
; they are loaded, even if not all levels load them.
ArtTile_ArtNem_WaterSurface           = $0400
ArtTile_ArtNem_Button                 = $0424
ArtTile_ArtNem_HorizSpike             = $042C
ArtTile_ArtNem_Spikes                 = $0434
ArtTile_ArtNem_DignlSprng             = $043C
ArtTile_ArtNem_LeverSpring            = $0440
ArtTile_ArtNem_VrtclSprng             = $045C
ArtTile_ArtNem_HrzntlSprng            = $0470
ArtTile_ArtNem_Spikes_S1              = $0578

; EHZ, HTZ
ArtTile_ArtKos_Checkers               = ArtTile_ArtKos_LevelArt+$0158
ArtTile_ArtUnc_Flowers1               = $0394
ArtTile_ArtUnc_Flowers2               = $0396
ArtTile_ArtUnc_Flowers3               = $0398
ArtTile_ArtUnc_Flowers4               = $039A
ArtTile_ArtNem_Buzzer                 = $03D2

; WFZ, SCZ
ArtTile_ArtNem_WfzHrzntlPrpllr        = $03CD
ArtTile_ArtNem_Clouds                 = $054F
ArtTile_ArtNem_WfzVrtclPrpllr         = $0561
ArtTile_ArtNem_Balkrie                = $0565

; ---------------------------------------------------------------------------
; Level-specific objects and badniks.

; EHZ
ArtTile_ArtUnc_EHZPulseBall           = $039C
ArtTile_ArtNem_Waterfall              = $039E
ArtTile_ArtNem_EHZ_Bridge             = $03B6
ArtTile_ArtNem_Buzzer_Fireball_2      = $03BE
ArtTile_ArtNem_Coconuts               = $03EE
ArtTile_ArtNem_Masher                 = $0414
ArtTile_ArtUnc_EHZMountains           = $0500

; MTZ
ArtTile_ArtNem_Crabmeat               = $031C
ArtTile_ArtUnc_Lava                   = $0340
ArtTile_ArtUnc_MTZCylinder            = $034C
ArtTile_ArtUnc_MTZAnimBack_1          = $035C
ArtTile_ArtUnc_MTZAnimBack_2          = $0362
ArtTile_ArtNem_MtzSupernova           = $0368
ArtTile_ArtNem_MtzWheel               = $0378
ArtTile_ArtNem_MtzWheelIndent         = $03F0
ArtTile_ArtNem_LavaCup                = $03F9
ArtTile_ArtNem_BoltEnd_Rope           = $03FD
ArtTile_ArtNem_MtzSteam               = $0405
ArtTile_ArtNem_MtzSpikeBlock          = $0414
ArtTile_ArtNem_MtzSpike               = $041C
ArtTile_ArtNem_MtzMantis              = $043C
ArtTile_ArtNem_MtzAsstBlocks          = $0500
ArtTile_ArtNem_MtzLavaBubble          = $0536
ArtTile_ArtNem_MtzCog                 = $055F
ArtTile_ArtNem_MtzUnkBlocks           = $056B

; WFZ
ArtTile_ArtNem_WfzScratch             = $0379
ArtTile_ArtNem_WfzTiltPlatforms       = $0393
ArtTile_ArtNem_WfzVrtclLazer          = $039F
ArtTile_ArtNem_WfzWallTurret          = $03AB
ArtTile_ArtNem_WfzHrzntlLazer         = $03C3
ArtTile_ArtNem_WfzConveyorBeltWheel   = $03EA
ArtTile_ArtNem_WfzHook                = $03FA
ArtTile_ArtNem_WfzHook_Fudge          = ArtTile_ArtNem_WfzHook + 4 ; Bad mappings...
ArtTile_ArtNem_WfzBeltPlatform        = $040E
ArtTile_ArtNem_WfzGunPlatform         = $041A
ArtTile_ArtNem_WfzUnusedBadnik        = $0450
ArtTile_ArtNem_WfzLaunchCatapult      = $045C
ArtTile_ArtNem_WfzSwitch              = $0461
ArtTile_ArtNem_WfzThrust              = $0465
ArtTile_ArtNem_WfzFloatingPlatform    = $046D
ArtTile_ArtNem_BreakPanels            = $05E8

; SCZ
ArtTile_ArtNem_Turtloid               = $038A
ArtTile_ArtNem_Nebula                 = $036E

; HTZ
ArtTile_ArtNem_Rexon                  = $037E
ArtTile_ArtNem_Buzzer_Fireball_1      = $039E
ArtTile_ArtNem_HtzRock                = $03B2
ArtTile_ArtNem_HtzSeeSaw              = $03C6
ArtTile_ArtNem_Sol                    = $03DE
ArtTile_ArtNem_HtzZipline             = $03E6
ArtTile_ArtNem_HtzFireball            = $0416
ArtTile_ArtNem_HtzValveBarrier        = $0426
ArtTile_ArtUnc_HTZMountains           = $0500
ArtTile_ArtUnc_HTZClouds              = ArtTile_ArtUnc_HTZMountains + $18
ArtTile_ArtNem_Spiker                 = $0520

; OOZ
ArtTile_ArtUnc_OOZPulseBall           = $02B6
ArtTile_ArtUnc_OOZSquareBall1         = $02BA
ArtTile_ArtUnc_OOZSquareBall2         = $02BE
ArtTile_ArtUnc_Oil1                   = $02C2
ArtTile_ArtUnc_Oil2                   = $02D2
ArtTile_ArtNem_OOZBurn                = $02E2
ArtTile_ArtNem_OOZElevator            = $02F4
ArtTile_ArtNem_SpikyThing             = $030C
ArtTile_ArtNem_BurnerLid              = $032C
ArtTile_ArtNem_StripedBlocksVert      = $0332
ArtTile_ArtNem_Oilfall                = $0336
ArtTile_ArtNem_Oilfall2               = $0346
ArtTile_ArtNem_BallThing              = $0354
ArtTile_ArtNem_LaunchBall             = $0368
ArtTile_ArtNem_OOZPlatform            = $039D
ArtTile_ArtNem_PushSpring             = $03C5
ArtTile_ArtNem_OOZSwingPlat           = $03E3
ArtTile_ArtNem_StripedBlocksHoriz     = $03FF
ArtTile_ArtNem_OOZFanHoriz            = $0403
ArtTile_ArtNem_Aquis                  = $0500
ArtTile_ArtNem_Octus                  = $0538

; MCZ
ArtTile_ArtNem_Flasher                = $03A8
ArtTile_ArtNem_Crawlton               = $03C0
ArtTile_ArtNem_Crate                  = $03D4
ArtTile_ArtNem_MCZCollapsePlat        = $03F4
ArtTile_ArtNem_VineSwitch             = $040E
ArtTile_ArtNem_VinePulley             = $041E
ArtTile_ArtNem_MCZGateLog             = $043C

; CNZ
ArtTile_ArtNem_Crawl                  = $0340
ArtTile_ArtNem_BigMovingBlock         = $036C
ArtTile_ArtNem_CNZSnake               = $037C
ArtTile_ArtNem_CNZBonusSpike          = $0380
ArtTile_ArtNem_CNZElevator            = $0384
ArtTile_ArtNem_CNZCage                = $0388
ArtTile_ArtNem_CNZHexBumper           = $0394
ArtTile_ArtNem_CNZRoundBumper         = $039A
ArtTile_ArtNem_CNZFlipper             = $03B2
ArtTile_ArtNem_CNZMiniBumper          = $03E6
ArtTile_ArtNem_CNZDiagPlunger         = $0402
ArtTile_ArtNem_CNZVertPlunger         = $0422

; Specific to 1p CNZ
ArtTile_ArtUnc_CNZFlipTiles_1         = $0330
ArtTile_ArtUnc_CNZFlipTiles_2         = $0540
ArtTile_ArtUnc_CNZSlotPics_1          = $0550
ArtTile_ArtUnc_CNZSlotPics_2          = $0560
ArtTile_ArtUnc_CNZSlotPics_3          = $0570

; Specific to 2p CNZ
ArtTile_ArtUnc_CNZFlipTiles_1_2p      = $0330
ArtTile_ArtUnc_CNZFlipTiles_2_2p      = $0740
ArtTile_ArtUnc_CNZSlotPics_1_2p       = $0750
ArtTile_ArtUnc_CNZSlotPics_2_2p       = $0760
ArtTile_ArtUnc_CNZSlotPics_3_2p       = $0770

; CPZ
ArtTile_ArtNem_ConstructionStripes_1  = $0328
ArtTile_ArtUnc_CPZAnimBack            = $0370
ArtTile_ArtNem_CPZMetalThings         = $0373
ArtTile_ArtNem_ConstructionStripes_2  = $0394
ArtTile_ArtNem_CPZBooster             = $039C
ArtTile_ArtNem_CPZElevator            = $03A0
ArtTile_ArtNem_CPZAnimatedBits        = $03B0
ArtTile_ArtNem_CPZTubeSpring          = $03E0
ArtTile_ArtNem_CPZStairBlock          = $0418
ArtTile_ArtNem_CPZMetalBlock          = $0430
ArtTile_ArtNem_CPZDroplet             = $043C
ArtTile_ArtNem_Grabber                = $0500
ArtTile_ArtNem_Spiny                  = $052D

; DEZ
ArtTile_ArtUnc_DEZAnimBack            = $0326

; ARZ
ArtTile_ArtNem_ARZBarrierThing        = $03F8
ArtTile_ArtNem_Leaves                 = $0410
ArtTile_ArtNem_ArrowAndShooter        = $0417
ArtTile_ArtUnc_Waterfall3             = $0428
ArtTile_ArtUnc_Waterfall2             = $042C
ArtTile_ArtUnc_Waterfall1_1           = $0430
ArtTile_ArtNem_Whisp                  = $0500
ArtTile_ArtNem_Grounder               = $0509
ArtTile_ArtNem_ChopChop               = $053B
ArtTile_ArtUnc_Waterfall1_2           = $0557
ArtTile_ArtNem_BigBubbles             = $055B

; MZ
ArtTile_ArtNem_Buzz_MZ                = $0298
ArtTile_ArtUnc_MzLava2                = $02D2
ArtTile_ArtUnc_MzLava1                = $02E2
ArtTile_ArtUnc_MzTorch                = $02F2
ArtTile_ArtNem_MzMetal                = $0300
ArtTile_ArtNem_MzFire                 = $0345
ArtTile_ArtNem_Swing                  = $0380
ArtTile_ArtNem_MzGlass                = $038E
ArtTile_ArtNem_Lava                   = $03A8
ArtTile_ArtNem_Cater_MZ               = $0444
ArtTile_ArtNem_MzBlock                = $0500
ArtTile_ArtNem_Basaran                = $0518
ArtTile_ArtNem_MzSwitch               = $0570

; SYZ
ArtTile_ArtNem_Bumper                 = $0380
ArtTile_ArtNem_SyzSpike1              = $0396
ArtTile_ArtNem_SyzSpike2              = $03BA
ArtTile_ArtNem_Crabmeat_SYZ           = $03BE
ArtTile_ArtNem_LampEmergency          = $03E7
ArtTile_ArtNem_Roller                 = $0402
ArtTile_ArtNem_Bumper2                = $0402
ArtTile_ArtNem_SyzLight               = $0418
ArtTile_ArtNem_LzSwitch               = $0454
ArtTile_ArtNem_Yadrin                 = $0500
ArtTile_ArtNem_Buzz_SYZ               = $053D

; LZ
ArtTile_ArtNem_LzBlock1               = $01E0
ArtTile_ArtNem_LzBlock2               = $01F0
ArtTile_ArtNem_LzSplash               = $0259
ArtTile_ArtNem_Gargoyle               = $02E9
ArtTile_ArtNem_LzWater                = $0300
ArtTile_ArtNem_LzSpikeBall            = $0310
ArtTile_ArtNem_FlapDoor               = $0328
ArtTile_ArtNem_LzBlock4               = $0339
ArtTile_ArtNem_Burrobot               = $0348
ArtTile_ArtNem_LzPlatfm               = $03A2
ArtTile_ArtNem_LzBlock3               = $03BC
ArtTile_ArtNem_LzDoor1                = $03C4
ArtTile_ArtNem_Harpoon                = $03CC
ArtTile_ArtNem_LzPole                 = $03DE
ArtTile_ArtNem_LzDoor2                = $03E6
ArtTile_ArtNem_LzWheel                = $03F6
ArtTile_ArtNem_Cork                   = $0500
ArtTile_ArtNem_Jaws                   = $0510
ArtTile_ArtNem_Orbinaut_LZ            = $0530
ArtTile_ArtNem_Spikes_LZ              = $0550
; HPZ
ArtTile_ArtUnc_HPZPulseOrb_1          = $02F0
ArtTile_ArtUnc_HPZPulseOrb_2          = $02F8
ArtTile_ArtUnc_HPZPulseOrb_3          = $0300
ArtTile_ArtNem_HPZ_Bridge             = $0308
ArtTile_ArtNem_HPZ_Waterfall          = $031D
ArtTile_ArtNem_HPZPlatform            = $0352
ArtTile_ArtNem_HPZOrb                 = $0362
ArtTile_Hpz_Unknow_Platform           = $0384
ArtTile_ArtNem_HPZ_Emerald            = $039A
ArtTile_ArtUnc_Big                    = $03BA
ArtTile_ArtNem_HPZBreakBlock          = $043C
ArtTile_Hpz_Dinobot                   = $0500
ArtTile_Hpz_Batbot                    = $0530
; ---------------------------------------------------------------------------
; Bosses
; Common tiles for some bosses (any for which no eggpod tiles are defined,
; except for WFZ and DEZ bosses).
ArtTile_ArtNem_Eggpod_4               = $0500
; Common tiles for all bosses.
ArtTile_ArtNem_FieryExplosion         = $0580

; CPZ boss
ArtTile_ArtNem_EggpodJets_1           = $0418
ArtTile_ArtNem_Eggpod_3               = $0420
ArtTile_ArtNem_CPZBoss                = $0500
ArtTile_ArtNem_BossSmoke_1            = $0570

; EHZ boss
ArtTile_ArtNem_Eggpod_1               = $03A0
ArtTile_ArtNem_EHZBoss                = $0400
ArtTile_ArtNem_EggChoppers            = $056C

; HTZ boss
ArtTile_ArtNem_Eggpod_2               = $03C1
ArtTile_ArtNem_HTZBoss                = $0421
ArtTile_ArtNem_BossSmoke_2            = $05E4

; ARZ boss
ArtTile_ArtNem_ARZBoss                = $03E0

; MCZ boss
ArtTile_ArtNem_MCZBoss                = $03C0
ArtTile_ArtUnc_FallingRocks           = $0560

; CNZ boss
ArtTile_ArtNem_CNZBoss                = $0407
ArtTile_ArtNem_CNZBoss_Fudge          = ArtTile_ArtNem_CNZBoss - $60 ; Badly reused mappings...

; MTZ boss
ArtTile_ArtNem_MTZBoss                = $037C
ArtTile_ArtNem_EggpodJets_2           = $0560

; OOZ boss
ArtTile_ArtNem_OOZBoss                = $038C

; WFZ and DEZ
ArtTile_ArtNem_RobotnikUpper          = $0500
ArtTile_ArtNem_RobotnikRunning        = $0518
ArtTile_ArtNem_RobotnikLower          = $0564

; WFZ boss
ArtTile_ArtNem_WFZBoss                = $0379

; DEZ
ArtTile_ArtNem_DEZBoss                = $0330
ArtTile_ArtNem_DEZWindow              = $0378
ArtTile_ArtNem_SilverSonic            = $0380
ArtTile_ArtNem_RobotnikUpper_DEZ      = ArtTile_ArtNem_RobotnikUpper - $1D8 ; Bad reused mappings...
; ---------------------------------------------------------------------------
; Universal locations.

; Animals.
ArtTile_ArtNem_Animal_1               = $0580
ArtTile_ArtNem_Animal_2               = $0594

; Game over.
ArtTile_ArtNem_Game_Over              = $04DE

; Titlecard.
ArtTile_ArtNem_TitleCard              = $0580
ArtTile_LevelName                     = $05DE

; End of level.
ArtTile_ArtNem_Signpost               = $0434
ArtTile_HUD_Bonus_Score               = $0520
ArtTile_ArtNem_Perfect                = $0540
ArtTile_ArtNem_ResultsText            = $05B0
ArtTile_ArtUnc_Signpost               = $05E8
ArtTile_ArtNem_MiniCharacter          = $05F4
ArtTile_ArtNem_Capsule                = $0680

; Tornado.
ArtTile_ArtNem_Tornado                = $0500
ArtTile_ArtNem_TornadoThruster        = $0561

; Some common objects; these are loaded on all aquatic levels.
ArtTile_ArtNem_Explosion              = $05A4
ArtTile_ArtNem_Bubbles                = $05E8
ArtTile_ArtNem_SuperSonic_stars       = $05F2

; Universal (used on all standard levels).
ArtTile_Emerald                       = $047C
ArtTile_EmeraldRadar                  = $0480
ArtTile_EmeraldRadarHPZ               = $03EC
ArtTile_EmeraldBonus                  = $048C
ArtTile_ArtNem_TailsDust              = $048C
ArtTile_ArtNem_SonicDust              = $049C
ArtTile_ArtNem_Numbers                = $04AC
ArtTile_ArtNem_Shield                 = $04BE
ArtTile_ArtNem_Invincible_stars       = $04DE
ArtTile_ArtNem_Powerups               = $0680
ArtTile_ArtNem_Ring                   = $06BC
ArtTile_ArtNem_HUD                    = ArtTile_ArtNem_Powerups + $4A
ArtTile_ArtUnc_Sonic                  = $0780
ArtTile_ArtUnc_Tails                  = $07A0
ArtTile_ArtUnc_Tails_Tails            = $07B0

; ---------------------------------------------------------------------------
; HUD. The HUD components are linked in a chain, and linked to
; power-ups, because the mappings of monitors and lives counter(s)
; depend on one another. If you want to alter these (for example,
; because you need the VRAM for something else), you will probably
; have to edit the mappings (or move the power-ups and HUD as a
; single block unit).
ArtTile_HUD_Score_Text                = ArtTile_ArtNem_HUD
ArtTile_HUD_Ring_Text                 = ArtTile_HUD_Score_Text + 6
ArtTile_HUD_Time_Text                 = ArtTile_HUD_Ring_Text + 8
ArtTile_HUD_Score_E                   = ArtTile_HUD_Time_Text + 8
ArtTile_HUD_Score                     = ArtTile_HUD_Score_E + 2
ArtTile_HUD_Score_0                   = ArtTile_HUD_Score + $C
ArtTile_HUD_Minutes                   = ArtTile_HUD_Score_0 + 2
ArtTile_HUD_Seconds                   = ArtTile_HUD_Minutes + 6
ArtTile_HUD_Rings                     = ArtTile_HUD_Seconds + 4
ArtTile_ArtUnc_2p_life_counter        = ArtTile_ArtNem_HUD + $2A
ArtTile_ArtUnc_2p_life_counter_lives  = ArtTile_ArtUnc_2p_life_counter + 9
ArtTile_ArtNem_life_counter           = ArtTile_ArtNem_HUD + $10A
ArtTile_ArtNem_life_counter_lives     = ArtTile_ArtNem_life_counter + 9

; ---------------------------------------------------------------------------
; Unused objects, objects with mappings never loaded, objects with
; missing mappings and/or tiles, objects whose mappings and/or tiles
; are never loaded.
ArtTile_ArtNem_UncPlatform            = $02B8
ArtTile_ArtNem_GHZ_Spiked_Log         = $0398
ArtTile_ArtNem_Unknown                = $03FA
ArtTile_ArtNem_BigRing                = $0400
ArtTile_ArtNem_FloatPlatform          = $0418
ArtTile_ArtNem_BigRing_Flash          = $0462
ArtTile_ArtNem_EndPoints              = $04B6
ArtTile_ArtNem_BreakWall              = $0590
ArtTile_ArtNem_Solid_Block            = $06C0

	codepage MenuText,STANDARD
	; set the character set for menu text
	charset '@',"\27\30\31\32\33\34\35\36\37\38\39\40\41\42\43\44\45\46\47\48\49\50\51\52\53\54\55"
	charset '0',"\16\17\18\19\20\21\22\23\24\25"
	charset '*',$1A
	charset ':',$1C
	charset '.',$1D
	charset '-',$6A
	charset '<',$6B
	charset '>',$6C
	charset $27,$6D ; '
	charset ' ',0

	codepage STANDARD

